from officialeye._internal.error.codes import ERR_SUPERVISION_CORRESPONDENCE_NOT_FOUND, ERR_SUPERVISION_INVALID_ENGINE_CONFIG
from officialeye._internal.error.error import OEError
from officialeye._internal.error.modules import ERR_MODULE_SUPERVISION


class ErrSupervision(OEError):

    def __init__(self, code: int, code_text: str, while_text: str, problem_text: str, /, *, is_regular: bool, **kwargs):
        super().__init__(ERR_MODULE_SUPERVISION, code, code_text, while_text, problem_text, is_regular=is_regular)


class ErrSupervisionCorrespondenceNotFound(ErrSupervision):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_SUPERVISION_CORRESPONDENCE_NOT_FOUND[0],
            ERR_SUPERVISION_CORRESPONDENCE_NOT_FOUND[1],
            while_text, problem_text, is_regular=True, **kwargs)


class ErrSupervisionInvalidEngineConfig(ErrSupervision):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_SUPERVISION_INVALID_ENGINE_CONFIG[0],
            ERR_SUPERVISION_INVALID_ENGINE_CONFIG[1],
            while_text, problem_text, is_regular=False, **kwargs)



from typing import Dict

from officialeye._internal.error.errors.template import ErrTemplateInvalidMutator
from officialeye._internal.mutation.mutator import Mutator
from officialeye._internal.mutation.mutators.clahe import CLAHEMutator
from officialeye._internal.mutation.mutators.grayscale import GrayscaleMutator
from officialeye._internal.mutation.mutators.non_local_means_denoising import NonLocalMeansDenoisingMutator
from officialeye._internal.mutation.mutators.rotate import RotateMutator


def load_mutator(mutator_id: str, config: Dict[str, any], /) -> Mutator:

    # TODO: make a container allowing one to dynamically load mutators (add such a container to OfficialEye's context).

    if mutator_id == GrayscaleMutator.MUTATOR_ID:
        return GrayscaleMutator(config)

    if mutator_id == NonLocalMeansDenoisingMutator.MUTATOR_ID:
        return NonLocalMeansDenoisingMutator(config)

    if mutator_id == CLAHEMutator.MUTATOR_ID:
        return CLAHEMutator(config)

    if mutator_id == RotateMutator.MUTATOR_ID:
        return RotateMutator(config)

    raise ErrTemplateInvalidMutator(
        f"while loading mutator '{mutator_id}'.",
        "Unknown mutator id."
    )


def load_mutator_from_dict(mutator_dict: Dict[str, any], /) -> Mutator:

    assert "id" in mutator_dict

    mutator_id = mutator_dict["id"]

    mutator_config = mutator_dict["config"] if "config" in mutator_dict else {}

    return load_mutator(mutator_id, mutator_config)



from typing import Dict

from officialeye._internal.diffobject.difference_modes import DIFF_MODE_ADD, DIFF_MODE_OVERRIDE, DIFF_MODE_REMOVE
from officialeye._internal.diffobject.exception import DiffObjectException
from officialeye._internal.diffobject.specification import DiffObjectSpecification
from officialeye._internal.diffobject.specification_entry import DiffObjectSpecificationEntry
from officialeye._internal.logger.singleton import get_logger


class DiffObjectExpansion:
    """
    This class maintains a state capturing the information gathered by applying partial objects of the same type in sequence.
    That is, this class allows you to build a complete specification-compliant object out of a sequence of partial objects,
    by applying the difference specifications contained in the partial objects appropriately.
    """

    def __init__(self, spec: DiffObjectSpecification, /):
        self._spec = spec
        self._cur_object = {}

    def add(self, partial_object: Dict[str, any], /):
        """
        Changes the current object by using the information from the given object.

        Arguments:
            partial_object: Which object to get the information from.

        Raises:
            DiffObjectException: In the event of a merge error.
        """

        def _do_add(specification_dict: Dict[str, any],
                    current_dict: Dict[str, any],
                    object_dict: Dict[str, any], /, *, previous_keys: str = ""):
            """
            Arguments:
                specification_dict: Current subdictionary of the specification represented as a dictionary
                current_dict: Corresponding subdictionary of `self._cur_object`
                object_dict: Corresponding subdictionary of `partial_object`
            """

            assert isinstance(specification_dict, dict)
            assert isinstance(current_dict, dict)
            assert isinstance(object_dict, dict)

            for key in specification_dict:

                if key not in object_dict:
                    # the current key is not present in the partial object at all
                    # hence, there is nothing to do
                    continue

                specification_entry = specification_dict[key]
                current_value = current_dict[key] if key in current_dict else None  # corresponding value in `self._cur_object`
                object_value = object_dict[key]  # corresponding value in `partial_object`
                object_value_diff_mode = object_dict[f"${key}"] if f"${key}" in object_dict else None

                full_key = f"{previous_keys}{key}"

                get_logger().debug_verbose(f"Key: '{full_key}' Specification value: {specification_entry} "
                                           f"Object value: {object_value} Current value: {current_value}")

                if isinstance(specification_entry, dict):
                    # the specification says that there is a nested dictionary at the present key.
                    # therefore, we need to recursively apply the diffs
                    if current_value is not None:
                        new_current_dict = current_value
                    else:
                        new_current_dict = {}
                        current_dict[key] = new_current_dict

                    _do_add(specification_entry, new_current_dict, object_value, previous_keys=f"{previous_keys}{key}.")
                    continue

                # handle non-recursive cases
                assert isinstance(specification_entry, DiffObjectSpecificationEntry)

                # test whether the partial object contains an entry explicitly specifying a difference node
                if object_value_diff_mode is not None:
                    if object_value_diff_mode not in (
                        DIFF_MODE_OVERRIDE,
                        DIFF_MODE_ADD,
                        DIFF_MODE_REMOVE
                    ):
                        raise DiffObjectException(f"Unknown difference mode specification '{object_value_diff_mode}' for key '{full_key}'.")
                    diff_mode = object_value_diff_mode
                else:
                    # default difference mode
                    diff_mode = DIFF_MODE_OVERRIDE

                current_dict[key] = specification_entry.apply_diff(current_value, object_value, diff_mode)

        _do_add(
            self._spec.get_spec_as_dict(),
            self._cur_object,
            partial_object
        )

    def get_current_partial_object(self) -> Dict[str, any]:
        return self._cur_object.copy()

    def get_full_object(self) -> Dict[str, any]:
        """
        Retrieves the full object with all the differences applied.

        Raises:
            DiffObjectException: If there is not enough information to build the entire object (i.e., not all keys are present).
        """

        def _verify_object_completeness(cur_obj_dict: Dict[str, any], spec_dict: Dict[str, any], /, *, previous_keys: str = ""):

            for key in spec_dict:
                spec_entry = spec_dict[key]

                full_key = f"{previous_keys}{key}"

                if key not in cur_obj_dict:
                    raise DiffObjectException(f"Could not resolve value for key '{full_key}'.")

                cur_obj_value = cur_obj_dict[key]

                if isinstance(spec_entry, dict):
                    # completeness should be checked recursively
                    assert isinstance(cur_obj_value, dict)
                    _verify_object_completeness(cur_obj_value, spec_entry, previous_keys=f"{previous_keys}{key}.")
                    continue

                assert isinstance(spec_entry, DiffObjectSpecificationEntry)

        _verify_object_completeness(self._cur_object, self._spec.get_spec_as_dict())

        return self._cur_object.copy()



from typing import Dict, Generator

import numpy as np
import z3

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.supervision import ErrSupervisionInvalidEngineConfig
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.match import Match
from officialeye._internal.matching.result import MatchingResult
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.supervision.supervisor import Supervisor


class OrthogonalRegressionSupervisor(Supervisor):

    ENGINE_ID = "orthogonal_regression"

    def __init__(self, context: Context, template_id: str, kmr: MatchingResult, /):
        super().__init__(context, OrthogonalRegressionSupervisor.ENGINE_ID, template_id, kmr)

        def _z3_timeout_preprocessor(v: any) -> int:

            v = int(v)

            if v < 1:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{OrthogonalRegressionSupervisor.ENGINE_ID}' supervisor.",
                    f"The `z3_timeout` value ({v}) cannot be negative or zero."
                )

            return v

        self.get_config().set_value_preprocessor("z3_timeout", _z3_timeout_preprocessor)

        self._z3_context = z3.Context()

        # create variables for components of the translation matrix
        self._transformation_matrix = np.array([
            [z3.Real("a", ctx=self._z3_context), z3.Real("b", ctx=self._z3_context)],
            [z3.Real("c", ctx=self._z3_context), z3.Real("d", ctx=self._z3_context)]
        ], dtype=z3.AstRef)

        # keys: matches (instances of Match)
        # values: z3 integer variables representing the errors for each match,
        # i.e., how consistent the match is with the affine transformation model
        self._match_error: Dict[Match, z3.ArithRef] = {}

        for match in self._kmr.get_matches():
            self._match_error[match] = z3.Real(f"e_{match.get_debug_identifier()}", ctx=self._z3_context)

    def _get_consistency_check(self, match: Match, delta: np.ndarray, delta_prime: np.ndarray, /) -> z3.AstRef:
        """
        Generates a z3 formula asserting the consistency of the match with the affine linear transformation model.
        Consistency does not mean ideal matching of coordinates; rather, the template position with the affine
        transformation applied to it, must roughly be equal the target position for consistency to hold
        In other words, targetpoint = M * (templatepoint - offset), where offset is a vector and M is a 2x2 matrix
        """

        template_point = match.get_original_template_point()

        assert delta.shape == (2,)
        assert delta_prime.shape == (2,)
        assert template_point.shape == (2,)

        translated_template_point = self._transformation_matrix @ (template_point - delta) + delta_prime
        translated_template_point_x, translated_template_point_y = translated_template_point

        target_point_x, target_point_y = match.get_target_point()

        return z3.And(
            translated_template_point_x - target_point_x <= self._match_error[match],
            translated_template_point_x - target_point_x >= -self._match_error[match],
            translated_template_point_y - target_point_y <= self._match_error[match],
            translated_template_point_y - target_point_y >= -self._match_error[match],
        )

    def _run(self) -> Generator[SupervisionResult, None, None]:

        for anchor_match in self._kmr.get_matches():
            delta = anchor_match.get_original_template_point()
            delta_prime = anchor_match.get_target_point()

            error_lower_bounds = z3.And(*(self._match_error[match] >= 0 for match in self._kmr.get_matches()), self._z3_context)
            total_error = z3.Sum(*(self._match_error[match] for match in self._kmr.get_matches()), self._z3_context)

            solver = z3.Optimize(ctx=self._z3_context)
            solver.set("timeout", self.get_config().get("z3_timeout", default=2500))

            solver.add(error_lower_bounds)

            for match in self._kmr.get_matches():
                solver.add(self._get_consistency_check(match, delta, delta_prime))

            solver.minimize(total_error)

            _result = solver.check()

            if _result == z3.unsat:
                get_logger().debug("Z3 returned unsat.")
                return

            if _result == z3.unknown:
                get_logger().debug("Z3 returned unknown.")
                return

            assert _result == z3.sat

            model = solver.model()

            evaluator = np.vectorize(lambda var: float(model.eval(var, model_completion=True).as_fraction()))  # noqa: B023

            # extract transformation matrix from model
            transformation_matrix = evaluator(self._transformation_matrix)

            _result = SupervisionResult(self.template_id, self._kmr, delta, delta_prime, transformation_matrix)

            get_logger().debug(f"Error: {_result.get_weighted_mse()}")

            yield _result



from typing import Dict

import strictyaml as yml

from officialeye._internal.diffobject.specification_entry import DiffObjectSpecificationEntry


class DiffObjectSpecification:
    """
    Class representing a complete specification of an object that supports partial definitions and difference application.
    """

    def __init__(self, specification: Dict[str, any], /, *, validate_specification: bool = True):
        self._spec = specification
        if validate_specification:
            self._validate_spec()

    def get_spec_as_dict(self) -> Dict[str, any]:
        return self._spec

    def _validate_spec(self, /):

        def _validator(entry_point: Dict[str, any], /):
            for key in entry_point:
                assert isinstance(key, str)

                value = entry_point[key]

                if isinstance(value, DiffObjectSpecificationEntry):
                    continue

                if isinstance(value, dict):
                    _validator(value)
                    continue

                raise AssertionError()

        _validator(self._spec)

    def get_schema(self) -> yml.Validator:

        def _mapper(entry_point: Dict[str, any], /) -> yml.Validator:

            mapped_dict = {}
            assert isinstance(entry_point, dict)

            for key in entry_point:
                assert isinstance(key, str)
                value = entry_point[key]

                if isinstance(value, DiffObjectSpecificationEntry):

                    # the value might be provided explicitly
                    mapped_dict[yml.Optional(key)] = value.get_schema()

                    # or it can be obtained from a previous value by applying a difference strategy
                    mapped_dict[yml.Optional("$" + key)] = yml.Str()

                    continue

                if isinstance(value, dict):
                    mapped_dict[yml.Optional(key)] = _mapper(value)
                    continue

                raise AssertionError()

            return yml.Map(mapped_dict)

        return _mapper(self._spec)



from typing import Union

import strictyaml as yml

from officialeye._internal.diffobject.difference_modes import DIFF_MODE_ADD, DIFF_MODE_OVERRIDE, DIFF_MODE_REMOVE
from officialeye._internal.diffobject.exception import DiffObjectException
from officialeye._internal.diffobject.specification_entry import DiffObjectSpecificationEntry


class IntegerSpecificationEntry(DiffObjectSpecificationEntry):

    def __init__(self, validator: yml.Validator, /):
        super().__init__(validator)

    def apply_diff(self, current_value: Union[int, None], diff_value: int, diff_mode: str) -> int:
        assert current_value is None or isinstance(current_value, int)
        assert isinstance(diff_value, int)

        if diff_mode == DIFF_MODE_OVERRIDE:
            return diff_value

        if current_value is None:
            raise DiffObjectException(f"Could not apply difference mode '{diff_mode}' because the previous value is not available.")

        if diff_mode == DIFF_MODE_ADD:
            return current_value + diff_value

        if diff_mode == DIFF_MODE_REMOVE:
            return current_value - diff_value

        raise DiffObjectException(f"The integer type is incompatible with difference mode specification '{diff_mode}'.")



from typing import List

import cv2
import numpy as np

from officialeye._internal.context.context import Context
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.template.region.keypoint import TemplateKeypoint


class SupervisionResultVisualizer:

    def __init__(self, context: Context, result: SupervisionResult, target: cv2.Mat):
        self._context = context
        self._result = result
        self._target = target

        # sorting is important, because we will be traversing the list multiple times
        # and the code relies on the identity of traverse orders
        self._relevant_keypoint_ids: List[str] = sorted(result.get_relevant_keypoint_ids())

        self._palette_width = max(self.get_template().get_keypoint(keypoint_id).w
                                  for keypoint_id in self._relevant_keypoint_ids)

    def get_template(self):
        return self._context.get_template(self._result.template_id)

    def get_padded_keypoint_image(self, keypoint: TemplateKeypoint) -> np.ndarray:

        assert keypoint.w <= self._palette_width

        keypoint_image = keypoint.to_image()

        padding_width = self._palette_width - keypoint.w
        assert padding_width >= 0

        if keypoint.w == self._palette_width:
            return keypoint_image

        assert padding_width > 0

        """
        white_pixel = np.array([[[0xff, 0xff, 0xff]]], dtype=keypoint_image.dtype)
        vertical_bar = np.repeat(white_pixel, keypoint.h, axis=0)
        padding = np.repeat(vertical_bar, padding_width, axis=1)
        """

        padding = np.full((keypoint.h, padding_width, 3), 0xff, dtype=np.uint8)
        keypoint_padded = np.concatenate((padding, keypoint_image), axis=1)

        assert keypoint_padded.shape[0] == keypoint.h
        assert keypoint_padded.shape[1] == self._palette_width

        return keypoint_padded

    def render(self) -> cv2.Mat:

        keypoints_palette = cv2.vconcat([
            self.get_padded_keypoint_image(self.get_template().get_keypoint(keypoint_id))
            for keypoint_id in self._relevant_keypoint_ids
        ])

        assert keypoints_palette.shape[1] == self._palette_width

        if keypoints_palette.shape[0] > self._target.shape[0]:
            # the height of the palette is greater than the height of the target image.
            # we pad the target image from below, to ensure that the height is at least the height of the palette
            padding_height = keypoints_palette.shape[0] - self._target.shape[0]
            padding = np.full((padding_height, self._target.shape[1], 3), 0xff, dtype=np.uint8)
            img = np.concatenate((self._target, padding), axis=0)
        else:
            img = self._target.copy()

        if keypoints_palette.shape[0] < self._target.shape[0]:
            # the height of the palette is less than the height of the target image
            # pad the palette from below, to ensure that the height is at least the height of the target image
            padding_height = self._target.shape[0] - keypoints_palette.shape[0]
            padding = np.full((padding_height, keypoints_palette.shape[1], 3), 0xff, dtype=np.uint8)
            keypoints_palette = np.concatenate((keypoints_palette, padding), axis=0)

        visualization = cv2.hconcat([keypoints_palette, img])

        for match in self._result.get_keypoint_matching_result().get_matches():
            # compute the location of the match in the palette part of the visualization
            match_palette_location = match.get_template_point()

            _padding_width = self._palette_width - self.get_template().get_keypoint(match.keypoint_id).w
            match_palette_location[0] += _padding_width

            for keypoint_id in self._relevant_keypoint_ids:
                if keypoint_id == match.keypoint_id:
                    break
                # in the palette, there is a keypoint above the keypoint corresponding to the current match.
                # therefore, we need to adjust the palette location.
                match_palette_location[1] += self.get_template().get_keypoint(keypoint_id).h

            # compute the location of the match in the target image part of the visualization
            match_target_location = match.get_target_point()
            match_target_location[0] += self._palette_width

            # draw a line visualizing the match
            # TODO: make this depend on the weight of the match
            match_color = (0, 0xff, 0)  # green

            visualization = cv2.line(visualization, match_palette_location, match_target_location, match_color, 2)

        return visualization



import cv2
import numpy as np

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.matching import ErrMatchingInvalidEngineConfig
from officialeye._internal.matching.match import Match
from officialeye._internal.matching.matcher import Matcher
from officialeye._internal.matching.result import MatchingResult

_FLANN_INDEX_KDTREE = 1


class SiftFlannMatcher(Matcher):

    ENGINE_ID = "sift_flann"

    def __init__(self, context: Context, template_id: str, img: cv2.Mat, /):
        super().__init__(context, SiftFlannMatcher.ENGINE_ID, template_id, img)

        def _preprocess_sensitivity(value: any) -> float:

            value = float(value)

            if value < 0.0:
                raise ErrMatchingInvalidEngineConfig(
                    f"while loading the '{SiftFlannMatcher.ENGINE_ID}' keypoint matcher",
                    f"The `sensitivity` value ({self._sensitivity}) cannot be negative."
                )

            if value > 1.0:
                raise ErrMatchingInvalidEngineConfig(
                    f"while loading the '{SiftFlannMatcher.ENGINE_ID}' keypoint matcher",
                    f"The `sensitivity` value ({self._sensitivity}) cannot exceed 1.0."
                )

            return value

        self.get_config().set_value_preprocessor("sensitivity", _preprocess_sensitivity)

        self._sensitivity = self.get_config().get("sensitivity", default=0.7)

        self._debug_images = []
        self._result = MatchingResult(self._context, template_id)

        # initialize the SIFT engine in CV2
        # noinspection PyUnresolvedReferences
        self._sift = cv2.SIFT_create()

        # pre-compute the sift keypoints in the target image
        self._keypoints_target, self._destination_target = self._sift.detectAndCompute(self._img, None)

    def match_keypoint(self, pattern: cv2.Mat, keypoint_id: str, /):

        pattern = cv2.cvtColor(pattern, cv2.COLOR_BGR2GRAY)

        keypoints_pattern, destination_pattern = self._sift.detectAndCompute(pattern, None)

        index_params = {
            "algorithm": _FLANN_INDEX_KDTREE,
            "trees": 5
        }

        search_params = {
            "checks": 50
        }

        flann = cv2.FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(destination_pattern, self._destination_target, k=2)

        # we need to draw only good matches, so create a mask
        matches_mask = [[0, 0] for _ in range(len(matches))]

        # filter matches
        for i, (m, n) in enumerate(matches):

            if m.distance >= self._sensitivity * n.distance:
                continue

            matches_mask[i] = [1, 0]

            pattern_point = keypoints_pattern[m.queryIdx].pt
            target_point = self._keypoints_target[m.trainIdx].pt

            # maybe one should consider rounding values here, instead of simply stripping the floating-point part
            pattern_point = np.array(pattern_point, dtype=int)
            target_point = np.array(target_point, dtype=int)

            match = Match(self._context, self.template_id, keypoint_id, pattern_point, target_point)
            match.set_score(self._sensitivity * n.distance - m.distance)

            self._result.add_match(match)

        if self._context.visualization_generation_enabled():

            # noinspection PyTypeChecker
            debug_image = cv2.drawMatchesKnn(
                pattern,
                keypoints_pattern,
                self._img,
                self._keypoints_target,
                matches,
                None,
                matchColor=(0, 0xff, 0),
                singlePointColor=(0xff, 0, 0),
                matchesMask=matches_mask,
                flags=cv2.DrawMatchesFlags_DEFAULT
            )

            self._context.export_image(debug_image, file_name=f"match_{keypoint_id}.png")

    def match_finish(self):
        return self._result



from queue import Queue
from threading import Thread
from typing import List, Tuple, Union

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.error import OEError
from officialeye._internal.error.errors.io import ErrIOInvalidImage
from officialeye._internal.error.errors.supervision import ErrSupervisionCorrespondenceNotFound
from officialeye._internal.error.errors.template import ErrTemplateInvalidConcurrencyConfig
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.template.template import Template


class AnalysisWorker(Thread):

    def __init__(self, worker_id: int, queue: Queue, target: cv2.Mat, /):
        Thread.__init__(self)

        self.worker_id = worker_id
        self.queue = queue

        self._target = target
        self._results: List[Tuple[Union[SupervisionResult, None], Union[OEError, None]]] = []

    def run(self):

        while True:
            template: Template = self.queue.get()

            try:
                _current_result = template.run_analysis(self._target)
                self._results.append((_current_result, None))
            except OEError as err:
                self._results.append((None, err))
            finally:
                self.queue.task_done()

    def get_successful_results(self):
        for result, error in self._results:
            if result is None or error is not None:
                continue
            yield result

    def get_errors(self):
        for _, error in self._results:
            if error is not None:
                yield error


def do_analyze(context: Context, target: cv2.Mat, templates: List[Template], /, *,
               num_workers: int, interpretation_target: Union[cv2.Mat, None] = None):

    if len(templates) == 0:
        # the program should be a noop if there are no templates provided
        return

    if interpretation_target is None:
        # if not specified, interpret the given image
        interpretation_target = target
    else:
        # there is a custom interpretation target specified.
        # it is essential that it has the same shape as the target image.
        # for this reason, we should verify this here
        if interpretation_target.shape != target.shape:
            raise ErrIOInvalidImage(
                "while making sure that the target image and the interpretation target images have the same shape.",
                f"The shapes mismatch. "
                f"The target image has shape {target.shape}, while the interpretation target image has shape {interpretation_target.shape}."
            )

    assert interpretation_target is not None
    assert num_workers is not None

    if num_workers < 1:
        raise ErrTemplateInvalidConcurrencyConfig(
            "while setting up workers for analyzing the target image.",
            f"The provided number of workers ({num_workers}) cannot be less than one."
        )

    if num_workers > 0xff:
        raise ErrTemplateInvalidConcurrencyConfig(
            "while setting up workers for analyzing the target image.",
            f"The provided number of workers ({num_workers}) is too high."
        )

    queue = Queue(maxsize=len(templates))

    workers = [AnalysisWorker(worker_id, queue, target) for worker_id in range(num_workers)]

    for worker in workers:
        worker.daemon = True
        worker.start()

    for template in templates:
        queue.put(template)

    queue.join()

    best_result = None
    best_result_score = -1.0

    # a list containing regular errors that occurred in workers
    regular_errors = []

    for worker in workers:
        for result in worker.get_successful_results():
            assert result is not None

            result_score = result.get_score()
            if result_score > best_result_score:
                best_result_score = result_score
                best_result = result

        for error in worker.get_errors():
            assert error is not None

            # we ignore regular errors here, because they may well be simply caused by trying to match
            # a given document against an invalid template, which is normal behavior
            if not error.is_regular:
                raise error
            else:
                get_logger().debug(f"Worker {worker.worker_id} returned the following non-regular error {error.code_text}:")
                get_logger().debug_oe_error(error)
                regular_errors.append(error)

    # note: best_result may be None here

    if best_result is None:

        error = ErrSupervisionCorrespondenceNotFound(
            "while running supervisor",
            "could not establish correspondence of the image with any of the templates provided"
        )

        for worker_error in regular_errors:
            error.add_cause(worker_error)

        raise error

    io_driver = context.get_io_driver()
    io_driver.handle_supervision_result(interpretation_target, best_result)



from typing import Dict

import cv2

from officialeye._internal.mutation.mutator import Mutator


class CLAHEMutator(Mutator):
    """
    Implementation of constrast increase via Contrast Limited Adaptive Histogram Equalization.
    """

    MUTATOR_ID = "clahe"

    def __init__(self, config: Dict[str, any], /):
        super().__init__(CLAHEMutator.MUTATOR_ID, config)

    def mutate(self, img: cv2.Mat, /) -> cv2.Mat:

        lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)
        l_channel, a, b = cv2.split(lab)

        # apply CLAHE to the L-channel
        # TODO: make parameters configurable
        clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))

        cl = clahe.apply(l_channel)

        # merge the CLAHE enhanced L-channel with the a and b channels
        limg = cv2.merge((cl, a, b))

        # convert back to BGR color space
        return cv2.cvtColor(limg, cv2.COLOR_LAB2BGR)



import abc
from typing import Callable, Dict


class Config(abc.ABC):

    def __init__(self, config_dict: Dict[str, any], /):
        self._config_dict = config_dict

        self._value_preprocessors: Dict[str, Callable[[str], any]] = {}

    def set_value_preprocessor(self, key: str, preprocessor: Callable[[str], any], /):
        self._value_preprocessors[key] = preprocessor

    @abc.abstractmethod
    def _get_invalid_key_error(self, key: str, /):
        raise NotImplementedError()

    def get(self, key: str, /, *, default=None):

        if key not in self._config_dict:

            if default is None:
                raise self._get_invalid_key_error(key)

            return default

        _value = self._config_dict[key]

        # apply value preprocessor
        if key in self._value_preprocessors:
            _value = self._value_preprocessors[key](_value)

        return _value



import os
import time
from typing import Dict, Generator, List, Union

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.io import ErrIOInvalidPath
from officialeye._internal.error.errors.template import (
    ErrTemplateInvalidFeature,
    ErrTemplateInvalidKeypoint,
    ErrTemplateInvalidMatchingEngine,
    ErrTemplateInvalidSupervisionEngine,
)
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.matcher import Matcher
from officialeye._internal.matching.matchers.sift_flann import SiftFlannMatcher
from officialeye._internal.matching.result import MatchingResult
from officialeye._internal.mutation.loader import load_mutator_from_dict
from officialeye._internal.mutation.mutator import Mutator
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.supervision.supervisors.combinatorial import CombinatorialSupervisor
from officialeye._internal.supervision.supervisors.least_squares_regression import LeastSquaresRegressionSupervisor
from officialeye._internal.supervision.supervisors.orthogonal_regression import OrthogonalRegressionSupervisor
from officialeye._internal.supervision.visualizer import SupervisionResultVisualizer
from officialeye._internal.template.feature_class.loader import load_template_feature_classes
from officialeye._internal.template.feature_class.manager import FeatureClassManager
from officialeye._internal.template.region.feature import TemplateFeature
from officialeye._internal.template.region.keypoint import TemplateKeypoint


class Template:

    def __init__(self, context: Context, yaml_dict: Dict[str, any], path_to_template: str, /):

        self._context = context

        self._path_to_template = path_to_template

        self.template_id = yaml_dict["id"]
        self.name = yaml_dict["name"]
        self._source = yaml_dict["source"]

        self.height, self.width, _ = self.load_source_image().shape

        self._source_mutators: List[Mutator] = [
            load_mutator_from_dict(mutator_dict) for mutator_dict in yaml_dict["mutators"]["source"]
        ]

        self._target_mutators: List[Mutator] = [
            load_mutator_from_dict(mutator_dict) for mutator_dict in yaml_dict["mutators"]["target"]
        ]

        self._keypoints: Dict[str, TemplateKeypoint] = {}
        self._features: Dict[str, TemplateFeature] = {}

        for keypoint_id in yaml_dict["keypoints"]:
            keypoint_dict = yaml_dict["keypoints"][keypoint_id]
            keypoint_dict["id"] = keypoint_id
            keypoint = TemplateKeypoint(self._context, self.template_id, keypoint_dict)

            if keypoint.region_id in self._keypoints:
                raise ErrTemplateInvalidKeypoint(
                    f"while initializing keypoint '{keypoint_id}' of template '{self.template_id}'",
                    f"There is already a keypoint with the same identifier '{keypoint.region_id}'."
                )

            if keypoint.region_id in self._features:
                raise ErrTemplateInvalidKeypoint(
                    f"while initializing keypoint '{keypoint_id}' of template '{self.template_id}'",
                    f"There is already a feature with the same identifier '{keypoint.region_id}'."
                )

            self._keypoints[keypoint.region_id] = keypoint

        self._matching = yaml_dict["matching"]
        self._supervision = yaml_dict["supervision"]

        # load feature classes
        self._feature_class_manager = load_template_feature_classes(self._context, yaml_dict["feature_classes"], self.template_id)

        # load features
        for feature_id in yaml_dict["features"]:
            feature_dict = yaml_dict["features"][feature_id]
            feature_dict["id"] = feature_id
            feature = TemplateFeature(self._context, self.template_id, feature_dict)

            if feature.region_id in self._keypoints:
                raise ErrTemplateInvalidFeature(
                    f"while initializing feature '{feature_id}' of template '{self.template_id}'",
                    f"There is already a keypoint with the same identifier '{feature.region_id}'."
                )

            if feature.region_id in self._features:
                raise ErrTemplateInvalidFeature(
                    f"while initializing feature '{feature_id}' of template '{self.template_id}'",
                    f"There is already a feature with the same identifier '{feature.region_id}'."
                )

            self._features[feature.region_id] = feature

        self._context.add_template(self)

    def validate(self):
        for feature in self.features():
            feature.validate_feature_class()

    def get_matching_engine(self) -> str:
        return self._matching["engine"]

    def get_supervision_engine(self) -> str:
        return self._supervision["engine"]

    def get_supervision_result(self) -> str:
        return self._supervision["result"]

    def get_supervision_config(self) -> dict:
        return self._supervision["config"]

    def get_matching_config(self) -> dict:
        matching_config = self._matching["config"]
        assert isinstance(matching_config, dict)
        return matching_config

    def get_feature_classes(self) -> FeatureClassManager:
        return self._feature_class_manager

    def _load_keypoint_matcher(self, target_img: cv2.Mat, /) -> Matcher:

        matching_engine = self.get_matching_engine()

        if matching_engine == SiftFlannMatcher.ENGINE_ID:
            return SiftFlannMatcher(self._context, self.template_id, target_img)

        raise ErrTemplateInvalidMatchingEngine(
            "while loading keypoint matcher",
            f"unknown matching engine '{matching_engine}'"
        )

    def features(self) -> Generator[TemplateFeature, None, None]:
        for feature_id in self._features:
            yield self._features[feature_id]

    def get_feature(self, feature_id: str, /) -> TemplateFeature:
        assert feature_id in self._features, "Invalid feature id"
        return self._features[feature_id]

    def keypoints(self) -> Generator[TemplateKeypoint, None, None]:
        for keypoint_id in self._keypoints:
            yield self._keypoints[keypoint_id]

    def get_keypoint(self, keypoint_id: str, /) -> TemplateKeypoint:
        assert keypoint_id in self._keypoints, "Invalid keypoint id"
        return self._keypoints[keypoint_id]

    def _get_source_image_path(self) -> str:
        if os.path.isabs(self._source):
            return self._source
        path_to_template_dir = os.path.dirname(self._path_to_template)
        path = os.path.join(path_to_template_dir, self._source)
        return os.path.normpath(path)

    def load_source_image(self) -> cv2.Mat:

        _image_path = self._get_source_image_path()

        if not os.path.isfile(_image_path):
            raise ErrIOInvalidPath(
                f"while loading template source image of template '{self.template_id}'.",
                f"Inferred path '{_image_path}' does not refer to a file."
            )

        if not os.access(_image_path, os.R_OK):
            raise ErrIOInvalidPath(
                f"while loading template source image of template '{self.template_id}'.",
                f"The file at path '{_image_path}' could not be read."
            )

        return cv2.imread(self._get_source_image_path(), cv2.IMREAD_COLOR)

    def _show(self, img: cv2.Mat, /, *, hide_features: bool, hide_keypoints: bool) -> cv2.Mat:

        if not hide_features:
            for feature in self.features():
                img = feature.visualize(img)

        if not hide_keypoints:
            for keypoint in self.keypoints():
                img = keypoint.visualize(img)

        return img

    def show(self, /, **kwargs) -> cv2.Mat:

        img = self.load_source_image()

        # apply template mutators to the target image
        for mutator in self._source_mutators:
            get_logger().debug(f"Applying mutator '{mutator.mutator_id}' to the source image of template '{self.template_id}'.")
            img = mutator.mutate(img)

        return self._show(img, **kwargs)

    def _load_supervisor(self, kmr: MatchingResult):
        superivision_engine = self.get_supervision_engine()

        if superivision_engine == LeastSquaresRegressionSupervisor.ENGINE_ID:
            return LeastSquaresRegressionSupervisor(self._context, self.template_id, kmr)

        if superivision_engine == OrthogonalRegressionSupervisor.ENGINE_ID:
            return OrthogonalRegressionSupervisor(self._context, self.template_id, kmr)

        if superivision_engine == CombinatorialSupervisor.ENGINE_ID:
            return CombinatorialSupervisor(self._context, self.template_id, kmr)

        raise ErrTemplateInvalidSupervisionEngine(
            "while loading supervisor",
            f"unknown supervision engine '{superivision_engine}'"
        )

    def run_analysis(self, target: cv2.Mat, /) -> Union[SupervisionResult, None]:
        # find all patterns in the target image

        _analysis_start_time = time.perf_counter(), time.process_time()

        # apply mutators to the target image
        for mutator in self._target_mutators:
            get_logger().debug(f"Applying mutator '{mutator.mutator_id}' to input image.")
            target = mutator.mutate(target)

        # start matching
        matcher = self._load_keypoint_matcher(target)

        for keypoint in self.keypoints():
            keypoint_pattern = keypoint.to_image()
            get_logger().debug(f"Running matcher for keypoint '{keypoint.region_id}'.")
            matcher.match_keypoint(keypoint_pattern, keypoint.region_id)

        keypoint_matching_result = matcher.match_finish()

        if self._context.visualization_generation_enabled():
            keypoint_matching_result.debug_print()

        keypoint_matching_result.validate()
        assert keypoint_matching_result.get_total_match_count() > 0

        _matching_ended_time = time.perf_counter(), time.process_time()

        get_logger().info(f"Matching succeeded in {_matching_ended_time[0] - _analysis_start_time[0]:.2f} seconds of real time "
                          f"and {_matching_ended_time[1] - _analysis_start_time[1]:.2f} seconds of CPU time.")

        # run supervision to obtain correspondence between template and target regions
        supervisor = self._load_supervisor(keypoint_matching_result)
        supervision_result = supervisor.run()

        if supervision_result is None:
            return None

        if self._context.visualization_generation_enabled():
            supervision_result_visualizer = SupervisionResultVisualizer(self._context, supervision_result, target)
            visualization = supervision_result_visualizer.render()
            self._context.export_image(visualization, file_name="matches.png")

        _supervision_ended_time = time.perf_counter(), time.process_time()

        get_logger().info(f"Supervision succeeded in {_supervision_ended_time[0] - _matching_ended_time[0]:.2f} seconds of real time "
                          f"and {_supervision_ended_time[1] - _matching_ended_time[1]:.2f} seconds of CPU time.")

        return supervision_result

    def __str__(self):
        return f"{self.name} ({self._source}, {len(self._keypoints)} keypoints, {len(self._features)} features)"



from typing import Dict

import cv2

from officialeye._internal.mutation.mutator import Mutator


class GrayscaleMutator(Mutator):

    MUTATOR_ID = "grayscale"

    def __init__(self, config: Dict[str, any], /):
        super().__init__(GrayscaleMutator.MUTATOR_ID, config)

    def mutate(self, img: cv2.Mat, /) -> cv2.Mat:
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)



from typing import Dict

from officialeye._internal.config.config import Config
from officialeye._internal.error.errors.supervision import ErrSupervisionInvalidEngineConfig


class SupervisorConfig(Config):

    def __init__(self, config_dict: Dict[str, any], supervision_engine_id: str, /):
        super().__init__(config_dict)

        self._supervision_engine_id = supervision_engine_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrSupervisionInvalidEngineConfig(
            f"while reading configuration of the '{self._supervision_engine_id}' supervision engine.",
            f"Could not find a value for key '{key}'."
        )



import tempfile
from typing import Dict

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.interpretation.method import InterpretationMethod
from officialeye._internal.interpretation.serializable import Serializable


class FileTempMethod(InterpretationMethod):

    METHOD_ID = "file_temp"

    def __init__(self, context: Context, config_dict: Dict[str, any]):
        super().__init__(context, FileTempMethod.METHOD_ID, config_dict)

        self._format = self.get_config().get("format", default="png")

    def interpret(self, feature_img: cv2.Mat, template_id: str, feature_id: str, /) -> Serializable:

        with tempfile.NamedTemporaryFile(prefix="officialeye_", suffix=f".{self._format}", delete=False) as fp:
            fp.close()

        cv2.imwrite(fp.name, feature_img)

        return fp.name



"""
Module containing the OfficialEye internal API.

WARNING!
Do not import it unless you know precisely what you are doing.
Instead, use the public API to interact with OfficialEye programatically.
"""


from officialeye._internal.error.codes import ERR_INTERNAL
from officialeye._internal.error.error import OEError
from officialeye._internal.error.modules import ERR_MODULE_INTERNAL


class ErrInternal(OEError):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(ERR_MODULE_INTERNAL, ERR_INTERNAL[0], ERR_INTERNAL[1], while_text, problem_text, is_regular=False)



import abc
from abc import ABC

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.error import OEError
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.template.template import Template


class IODriver(ABC):

    def __init__(self, context: Context):
        self._context = context

    @abc.abstractmethod
    def handle_supervision_result(self, target: cv2.Mat, result: SupervisionResult, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_show_result(self, template: Template, img: cv2.Mat, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_error(self, error: OEError, /):
        raise NotImplementedError()



import abc
from abc import ABC

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.matcher_config import KeypointMatcherConfig
from officialeye._internal.matching.result import MatchingResult


class Matcher(ABC):
    # TODO: migrate matcher to a separate module

    def __init__(self, context: Context, engine_id: str, template_id: str, img: cv2.Mat, /):
        super().__init__()

        self._context = context
        self._engine_id = engine_id

        self.template_id = template_id

        # retreive configurations for all keypoint matching engines
        matching_config = self.get_template().get_matching_config()

        assert isinstance(matching_config, dict)

        # get the configuration for the particular engine of interest
        if self._engine_id in matching_config:
            config_dict = matching_config[self._engine_id]
        else:
            get_logger().warn(
                self._context,
                f"Could not find any configuration entries for the '{self._engine_id}' matching engine that is being used."
            )
            config_dict = {}

        self._config = KeypointMatcherConfig(config_dict, engine_id)

        self._img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    @abc.abstractmethod
    def match_keypoint(self, pattern: cv2.Mat, keypoint_id: str, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def match_finish(self) -> MatchingResult:
        raise NotImplementedError()

    def get_template(self):
        return self._context.get_template(self.template_id)

    def get_config(self) -> KeypointMatcherConfig:
        return self._config



import strictyaml as yml

from officialeye._internal.diffobject.specification import DiffObjectSpecification
from officialeye._internal.diffobject.specification_entries.boolean import BooleanSpecificationEntry
from officialeye._internal.diffobject.specification_entries.list import ListSpecificationEntry
from officialeye._internal.diffobject.specification_entries.object import ObjectSpecificationEntry
from officialeye._internal.diffobject.specification_entries.string import StringSpecificationEntry

_alphanumeric_id_validator = yml.Regex(r"^[a-zA-Z0-9_]{1,64}$")

_mutator_specification = yml.Map({
    "id": _alphanumeric_id_validator,
    yml.Optional("config"): yml.EmptyDict() | yml.MapPattern(_alphanumeric_id_validator, yml.Any())
})

feature_class_object_specification = DiffObjectSpecification({
    "abstract": BooleanSpecificationEntry(yml.Bool()),
    "inherits": StringSpecificationEntry(_alphanumeric_id_validator),
    "mutators": ListSpecificationEntry(yml.EmptyList() | yml.Seq(_mutator_specification)),
    "interpretation": {
        "method": StringSpecificationEntry(_alphanumeric_id_validator),
        "config": ObjectSpecificationEntry(yml.EmptyDict() | yml.MapPattern(_alphanumeric_id_validator, yml.Any()))
    }
})


def generate_template_schema() -> yml.Map:
    global _alphanumeric_id_validator

    _keypoint_validator = yml.Map({
        "x": yml.Int(),
        "y": yml.Int(),
        "w": yml.Int(),
        "h": yml.Int(),
        "matches": yml.Map({
            "min": yml.Int(),
            "max": yml.Int()
        })
    })

    _oe_template_schema_feature_validator = yml.Map({
        "x": yml.Int(),
        "y": yml.Int(),
        "w": yml.Int(),
        "h": yml.Int(),
        yml.Optional("class"): _alphanumeric_id_validator
    })

    _feature_class_validator = feature_class_object_specification.get_schema()

    return yml.Map({
        "id": _alphanumeric_id_validator,
        "name": yml.Regex(r"^[a-zA-Z0-9_ ']{1,64}$"),
        "source": yml.Str(),
        "mutators": yml.Map({
            "source": yml.EmptyList() | yml.Seq(_mutator_specification),
            "target": yml.EmptyList() | yml.Seq(_mutator_specification)
        }),
        "keypoints": yml.MapPattern(_alphanumeric_id_validator, _keypoint_validator),
        "matching": yml.Map({
            "engine": _alphanumeric_id_validator,
            "config": yml.EmptyDict() | yml.MapPattern(
                _alphanumeric_id_validator,
                yml.EmptyDict() | yml.MapPattern(_alphanumeric_id_validator, yml.Any())
            )
        }),
        "supervision": yml.Map({
            "engine": _alphanumeric_id_validator,
            "config": yml.EmptyDict() | yml.MapPattern(
                _alphanumeric_id_validator,
                yml.EmptyDict() | yml.MapPattern(_alphanumeric_id_validator, yml.Any())
            ),
            "result": yml.Regex(r"^(first|random|best_mse|best_score)$")
        }),
        "feature_classes": yml.MapPattern(_alphanumeric_id_validator, _feature_class_validator),
        "features": yml.MapPattern(_alphanumeric_id_validator, _oe_template_schema_feature_validator)
    })



from typing import Dict

import cv2

from officialeye._internal.error.errors.template import ErrTemplateInvalidMutator
from officialeye._internal.mutation.mutator import Mutator


class NonLocalMeansDenoisingMutator(Mutator):

    MUTATOR_ID = "non_local_means_denoising"

    def __init__(self, config: Dict[str, any], /):
        super().__init__(NonLocalMeansDenoisingMutator.MUTATOR_ID, config)

        # setup configuration loading

        self.get_config().set_value_preprocessor("colored", bool)
        self.get_config().set_value_preprocessor("h", int)
        self.get_config().set_value_preprocessor("hForColorComponents", int)
        self.get_config().set_value_preprocessor("templateWindowSize", int)
        self.get_config().set_value_preprocessor("searchWindowSize", int)

        # load data from configuration

        self._colored_mode = self.get_config().get("colored", default=True)

        self._conf_h = self.get_config().get("h", default=10)
        self._conf_hForColorComponents = self.get_config().get("hForColorComponents", default=10)
        self._conf_templateWindowSize = self.get_config().get("templateWindowSize", default=7)
        self._conf_searchWindowSize = self.get_config().get("searchWindowSize", default=21)

        # validate templateWindowSize
        if self._conf_templateWindowSize < 1:
            raise ErrTemplateInvalidMutator(
                f"while loading mutator '{self.mutator_id}'.",
                f"The 'templateWindowSize' parameter must be positive, got '{self._conf_templateWindowSize}'."
            )

        if self._conf_templateWindowSize % 2 == 0:
            raise ErrTemplateInvalidMutator(
                f"while loading mutator '{self.mutator_id}'.",
                f"The 'templateWindowSize' parameter must be odd, got '{self._conf_templateWindowSize}'."
            )

        # validate searchWindowSize
        if self._conf_searchWindowSize < 1:
            raise ErrTemplateInvalidMutator(
                f"while loading mutator '{self.mutator_id}'.",
                f"The 'searchWindowSize' parameter must be positive, got '{self._conf_searchWindowSize}'."
            )

        if self._conf_searchWindowSize % 2 == 0:
            raise ErrTemplateInvalidMutator(
                f"while loading mutator '{self.mutator_id}'.",
                f"The 'searchWindowSize' parameter must be odd, got '{self._conf_searchWindowSize}'."
            )

    def mutate(self, img: cv2.Mat, /) -> cv2.Mat:

        if self._colored_mode:
            return cv2.fastNlMeansDenoisingColored(
                img,
                None,
                self._conf_h,
                self._conf_hForColorComponents,
                self._conf_templateWindowSize,
                self._conf_searchWindowSize
            )

        return cv2.fastNlMeansDenoising(
            img,
            None,
            self._conf_h,
            self._conf_templateWindowSize,
            self._conf_searchWindowSize
        )



from typing import List


class OEError(Exception):
    """
    Base class for all officialeye-related errors.
    """

    def __init__(self, module: str, code: int, code_text: str, while_text: str, problem_text: str, /, *,
                 is_regular: bool = False):
        super().__init__()

        assert code != 0

        self.code = code
        self.module = module
        self.code_text = code_text
        self.while_text = while_text
        self.problem_text = problem_text

        # regular errors are those that can happen if the tool is configured correctly, but has been provided invalid input
        # here, 'input' should be narrowly understood as end user input, not as the tool users' input
        # an example of such an input would be the image provided as a basis for the analysis
        # on the other hand, the template configuration file does not count as end user input
        self.is_regular = is_regular

        self._causes: List[OEError] = []
        self._external_causes: List[BaseException] = []

    def add_cause(self, cause, /):
        assert isinstance(cause, OEError)
        self._causes.append(cause)

    def get_causes(self):
        return self._causes

    def add_external_cause(self, cause: BaseException, /):
        self._external_causes.append(cause)

    def get_external_causes(self) -> List[BaseException]:
        return self._external_causes

    def serialize(self) -> dict:

        causes = [
            cause.serialize()
            for cause in self._causes
        ]

        return {
            "code": self.code,
            "code_text": self.code_text,
            "module": self.module,
            "while_text": self.while_text,
            "problem_text": self.problem_text,
            "is_regular": self.is_regular,
            "causes": causes,
            "external_causes": [
                str(ec) for ec in self._external_causes
            ]
        }



"""
OfficialEye main entry point.
"""

from typing import List, Union

import click
import cv2

from officialeye._internal.context.manager import ContextManager
from officialeye._internal.io.drivers.run import RunIODriver
from officialeye._internal.io.drivers.test import TestIODriver
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.template.analyze import do_analyze
from officialeye._internal.template.create import create_example_template_config_file
from officialeye._internal.template.schema.loader import load_template
from officialeye.meta import OFFICIALEYE_GITHUB, OFFICIALEYE_VERSION

_context_manager: ContextManager = ContextManager()


@click.group()
@click.option("-d", "--debug", is_flag=True, show_default=True, default=False, help="Enable debug mode.")
@click.option("--edir", type=click.Path(exists=True, file_okay=True, readable=True), help="Specify export directory.")
@click.option("-q", "--quiet", is_flag=True, show_default=True, default=False, help="Disable standard output messages.")
@click.option("-v", "--verbose", is_flag=True, show_default=True, default=False, help="Enable verbose logging.")
@click.option("-dl", "--disable-logo", is_flag=True, show_default=True, default=False, help="Disable the officialeye logo.")
@click.option("-re", "--raw-errors", is_flag=True, show_default=False, default=False, help="Do not handle errors.")
def cli(debug: bool, edir: str, quiet: bool, verbose: bool, disable_logo: bool, raw_errors: bool):
    global _context_manager

    # configure logger
    get_logger().debug_mode = debug
    get_logger().quiet_mode = quiet
    get_logger().verbose_mode = verbose
    get_logger().disable_logo = disable_logo

    # print OfficialEye logo if necessary
    get_logger().logo()

    # configure context manager
    if edir is not None:
        _context_manager.export_directory = edir

    if raw_errors:
        get_logger().warn("Raw error mode enabled. Use this mode only if you know precisely what you are doing!")
        _context_manager.handle_exceptions = False

    # print preliminary warning if necessary
    if get_logger().debug_mode:
        get_logger().warn("Debug mode enabled. Disable for production use to improve performance.")


# noinspection PyShadowingBuiltins
@click.command()
@click.argument("template_path", type=click.Path(exists=False, file_okay=True, readable=True, writable=True))
@click.argument("template_image", type=click.Path(exists=True, file_okay=True, readable=True, writable=False))
@click.option("--id", type=str, show_default=False, default="example", help="Specify the template identifier.")
@click.option("--name", type=str, show_default=False, default="Example", help="Specify the template name.")
@click.option("--force", is_flag=True, show_default=True, default=False, help="Create missing directories and overwrite file.")
def create(template_path: str, template_image: str, id: str, name: str, force: bool):
    """Creates a new template configuration file at the specified path."""
    create_example_template_config_file(template_path, template_image, id, name, force)


@click.command()
@click.argument("template_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.option("--hide-features", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of features.")
@click.option("--hide-keypoints", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of keypoints.")
def show(template_path: str, hide_features: bool, hide_keypoints: bool):
    """Exports template as an image with features visualized."""

    global _context_manager

    with _context_manager as oe_context:
        # setup IO driver
        oe_context.set_io_driver(TestIODriver(oe_context))

        # load template
        template = load_template(oe_context, template_path)

        # render resulting image
        img = template.show(hide_features=hide_features, hide_keypoints=hide_keypoints)

        # show rendered image
        oe_context.get_io_driver().handle_show_result(template, img)


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--workers", type=int, default=4, show_default=True, help="Specify number of threads to use for the pool of workers.")
@click.option("--interpret", type=click.Path(exists=True, file_okay=True, readable=True),
              default=None, help="Use the image at the specified path to run the interpretation phase.")
@click.option("--show-features", is_flag=True, show_default=False, default=False, help="Visualize the locations of features.")
@click.option("--visualize", is_flag=True, show_default=False, default=False, help="Generate visualizations of intermediate steps.")
def test(target_path: str, template_paths: List[str], workers: int, interpret: Union[str, None], show_features: bool, visualize: bool):
    """Visualizes the analysis of an image using one or more templates."""

    global _context_manager

    _context_manager.visualization_generation = visualize

    if _context_manager.visualization_generation:
        get_logger().warn("Visualization generation mode enabled. Disable for production use to improve performance.")

    with (_context_manager as oe_context):

        # setup IO driver
        io_driver = TestIODriver(oe_context)
        io_driver.visualize_features = show_features
        oe_context.set_io_driver(io_driver)

        # load target image
        target = cv2.imread(target_path, cv2.IMREAD_COLOR)

        # load interpretation target image if necessary
        interpretation_target: Union[cv2.Mat, None] = \
            None if interpret is None else cv2.imread(interpret, cv2.IMREAD_COLOR)

        # load templates
        templates = [load_template(oe_context, template_path) for template_path in template_paths]

        # perform analysis
        do_analyze(oe_context, target, templates, num_workers=workers, interpretation_target=interpretation_target)


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--workers", type=int, default=4, show_default=True, help="Specify number of threads to use for the pool of workers.")
@click.option("--interpret", type=click.Path(exists=True, file_okay=True, readable=True),
              default=None, help="Use the image at the specified path to run the interpretation phase.")
@click.option("--visualize", is_flag=True, show_default=False, default=False, help="Generate visualizations of intermediate steps.")
def run(target_path: str, template_paths: List[str], workers: int, interpret: Union[str, None], visualize: bool):
    """Applies one or more templates to an image."""

    global _context_manager

    _context_manager.visualization_generation = visualize

    if _context_manager.visualization_generation:
        get_logger().warn("Visualization generation mode enabled. Disable for production use to improve performance.")

    with _context_manager as oe_context:
        # setup IO driver
        oe_context.set_io_driver(RunIODriver(oe_context))

        # load target image
        target = cv2.imread(target_path, cv2.IMREAD_COLOR)

        # load interpretation target image if necessary
        interpretation_target: Union[cv2.Mat, None] = \
            None if interpret is None else cv2.imread(interpret, cv2.IMREAD_COLOR)

        # load templates
        templates = [load_template(oe_context, template_path) for template_path in template_paths]

        # perform analysis
        do_analyze(oe_context, target, templates, num_workers=workers, interpretation_target=interpretation_target)


@click.command()
def homepage():
    """Go to the officialeye's official GitHub homepage."""
    get_logger().info(f"Opening {OFFICIALEYE_GITHUB}")
    click.launch(OFFICIALEYE_GITHUB)


@click.command()
def version():
    """Print the version of OfficialEye."""
    get_logger().info(f"Version: {OFFICIALEYE_VERSION}")


cli.add_command(create)
cli.add_command(show)
cli.add_command(test)
cli.add_command(run)
cli.add_command(homepage)
cli.add_command(version)

if __name__ == "__main__":
    cli()



"""
A difference object is either a dictionary conforming to some schema, or a dictionary-like representation of how to change another dictionary.
This module contains everything related to difference objects.
"""


import json
import sys

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.error import OEError
from officialeye._internal.error.errors.io import ErrIOOperationNotSupportedByDriver
from officialeye._internal.io.driver import IODriver
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.template.template import Template


def _output_dict(d: dict):
    json.dump(d, sys.stdout, indent=4, ensure_ascii=False)
    sys.stdout.write("\n")
    sys.stdout.flush()


class RunIODriver(IODriver):

    def __init__(self, context: Context, /):
        super().__init__(context)

    def handle_show_result(self, template: Template, img: cv2.Mat, /):
        raise ErrIOOperationNotSupportedByDriver(
            f"while trying to output the result of showing the template '{template.template_id}'",
            "Driver 'run' does not support this operation."
        )

    def handle_error(self, error: OEError, /):
        _output_dict({
            "ok": False,
            "err": error.serialize()
        })

    def handle_supervision_result(self, target: cv2.Mat, result: SupervisionResult, /):

        assert result is not None

        template = self._context.get_template(result.template_id)

        feature_interpretation_dict = {}

        # extract the features from the target image
        for feature in template.features():

            feature_class = feature.get_feature_class()

            if feature_class is None:
                continue

            feature_img = result.get_feature_warped_region(target, feature)
            feature_img_mutated = feature.apply_mutators_to_image(feature_img)
            interpretation = feature.interpret_image(feature_img_mutated)

            feature_interpretation_dict[feature.region_id] = interpretation

        _output_dict({
            "ok": True,
            "template": result.template_id,
            "score": result.get_score(),
            "features": feature_interpretation_dict
        })



from typing import Generator

import numpy as np

from officialeye._internal.context.context import Context
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.result import MatchingResult
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.supervision.supervisor import Supervisor

_IND_A = 0
_IND_B = 1
_IND_C = 2
_IND_D = 3


class LeastSquaresRegressionSupervisor(Supervisor):

    ENGINE_ID = "least_squares_regression"

    def __init__(self, context: Context, template_id: str, kmr: MatchingResult, /):
        super().__init__(context, LeastSquaresRegressionSupervisor.ENGINE_ID, template_id, kmr)

    def _run(self) -> Generator[SupervisionResult, None, None]:

        match_count = self._kmr.get_total_match_count()

        for anchor_match in self._kmr.get_matches():
            delta = anchor_match.get_original_template_point()
            delta_prime = anchor_match.get_target_point()

            matrix = np.zeros((match_count << 1, 4), dtype=np.float64)
            rhs = np.zeros(match_count << 1, dtype=np.float64)

            for i, match in enumerate(self._kmr.get_matches()):
                first_constraint_id = i << 1
                second_constraint_id = first_constraint_id + 1

                s = match.get_original_template_point()
                d = match.get_target_point()

                matrix[first_constraint_id][_IND_A] = s[0] - delta[0]
                matrix[first_constraint_id][_IND_B] = s[1] - delta[1]
                rhs[first_constraint_id] = d[0] - delta_prime[0]

                matrix[second_constraint_id][_IND_C] = s[0] - delta[0]
                matrix[second_constraint_id][_IND_D] = s[1] - delta[1]
                rhs[second_constraint_id] = d[1] - delta_prime[1]

            regression_matrix = matrix.T @ matrix
            regression_matrix = np.linalg.inv(regression_matrix)
            rhs_applied = matrix.T @ rhs
            x = regression_matrix @ rhs_applied

            transformation_matrix = np.array([
                [x[_IND_A], x[_IND_B]],
                [x[_IND_C], x[_IND_D]]
            ])

            _result = SupervisionResult(self.template_id, self._kmr, delta, delta_prime, transformation_matrix)

            get_logger().debug(f"Current MSE: {_result.get_weighted_mse()}")

            yield _result



from types import TracebackType
from typing import Union

from officialeye._internal.context.context import Context
from officialeye._internal.error.error import OEError
from officialeye._internal.error.errors.internal import ErrInternal


class ContextManager:

    def __init__(self, /, *, handle_exceptions: bool = True, visualization_generation: bool = False,
                 export_directory: Union[str, None] = None):

        self._context: Union[Context, None] = None

        self.handle_exceptions = handle_exceptions

        self.visualization_generation = visualization_generation

        self.export_directory = export_directory

    def __enter__(self) -> Context:

        if self._context is not None:
            raise ErrInternal(
                "while entering an officialeye context.",
                "The present context manager has already got an associated context. Are you trying to reuse the context manager?"
            )

        self._context = Context(self, visualization_generation=self.visualization_generation)

        return self._context

    def __exit__(self, exception_type: any, exception_value: BaseException, traceback: TracebackType):

        if self._context is None:
            raise ErrInternal(
                "while leaving an officialeye context.",
                "The present context manager has no context associated with it."
            )

        self._context.dispose()

        if not self.handle_exceptions:
            return

        # handle the possible exception
        if exception_value is None:
            # there is no exception, nothing to handle
            return

        if isinstance(exception_value, OEError):
            oe_error = exception_value
        elif isinstance(exception_value, BaseException):
            oe_error = ErrInternal(
                "while leaving an officialeye context.",
                "An internal error occurred.",
            )
            oe_error.add_external_cause(exception_value)
        else:
            raise AssertionError()

        self._context.get_io_driver().handle_error(oe_error)

        # tell python that we have handled the exception ourselves
        return True



import abc
from abc import ABC

import strictyaml as yml


class DiffObjectSpecificationEntry(ABC):

    def __init__(self, validator: yml.Validator, /):
        self._validator = validator

    def get_schema(self) -> yml.Validator:
        """ Retrieves a schema for the current specification entry. """
        return self._validator

    @abc.abstractmethod
    def apply_diff(self, current_value: any, diff_value: any, diff_mode: str) -> any:
        """
        Calculates a new value for an entry, based on currently available value, some provided value, and a way of combining them.

        Arguments:
            current_value: The current value assigned to the entry, or None if there is no such value.
            diff_value: A new value provided for this entry.
            diff_mode: How the two values are to be combined to form a new value.

        Returns:
            The new value obtained by combining current_value with diff_value using diff_mode.

        Raises:
            DiffObjectException: If the difference mode cannot be applied to the present entry, for example, due to its type.
        """
        raise NotImplementedError()



from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidFeatureClass
from officialeye._internal.template.feature_class.manager import FeatureClassManager


def load_template_feature_classes(context: Context, feature_classes_dict: dict, template_id: str, /) -> FeatureClassManager:

    assert isinstance(feature_classes_dict, dict)

    _manager = FeatureClassManager(context, template_id)

    for class_id in feature_classes_dict:

        if _manager.contains_class(class_id):
            raise ErrTemplateInvalidFeatureClass(
                f"while loading feature classes of template '{template_id}'.",
                f"Class '{class_id}' has been defined more than once."
            )

        _manager.add_class(class_id, feature_classes_dict[class_id])

    _manager.inline_all_classes()

    return _manager



from typing import Dict

from officialeye._internal.context.context import Context
from officialeye._internal.diffobject.exception import DiffObjectException
from officialeye._internal.error.errors.template import ErrTemplateInvalidFeatureClass
from officialeye._internal.template.feature_class.const import IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID
from officialeye._internal.template.feature_class.feature_class import FeatureClass


class FeatureClassManager:

    def __init__(self, context: Context, template_id: str, /):
        self._context = context
        self._template_id = template_id
        self._classes: Dict[str, FeatureClass] = {
            IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID: FeatureClass(self, IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID, {
                "abstract": True
            })
        }

    def get_global_base_class(self) -> FeatureClass:
        return self._classes[IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID]

    def get_class(self, class_id: str, /):
        assert class_id in self._classes
        return self._classes[class_id]

    def get_template(self):
        return self._context.get_template(self._template_id)

    def contains_class(self, class_id: str, /) -> bool:
        return class_id in self._classes

    def add_class(self, class_id: str, class_dict: Dict[str, any], /):
        assert class_id not in self._classes
        self._classes[class_id] = FeatureClass(self, class_id, class_dict)

    def inline_all_classes(self):

        try:
            for class_id in self._classes:
                self._classes[class_id].inline()
        except DiffObjectException as err:
            raise ErrTemplateInvalidFeatureClass(
                f"while loading feature classes of template '{self._template_id}'.",
                err.problem
            ) from err



from officialeye._internal.error.codes import (
    ERR_TEMPLATE_ID_NOT_UNIQUE,
    ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG,
    ERR_TEMPLATE_INVALID_FEATURE,
    ERR_TEMPLATE_INVALID_FEATURE_CLASS,
    ERR_TEMPLATE_INVALID_INTERPRETATION,
    ERR_TEMPLATE_INVALID_KEYPOINT,
    ERR_TEMPLATE_INVALID_MATCHING_ENGINE,
    ERR_TEMPLATE_INVALID_MUTATOR,
    ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE,
    ERR_TEMPLATE_INVALID_SYNTAX,
)
from officialeye._internal.error.error import OEError
from officialeye._internal.error.modules import ERR_MODULE_TEMPLATE


class ErrTemplate(OEError):

    def __init__(self, code: int, code_text: str, while_text: str, problem_text: str, /, *, is_regular: bool = False, **kwargs):
        super().__init__(ERR_MODULE_TEMPLATE, code, code_text, while_text, problem_text, is_regular=is_regular)


class ErrTemplateInvalidSupervisionEngine(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE[0], ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidMatchingEngine(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_MATCHING_ENGINE[0], ERR_TEMPLATE_INVALID_MATCHING_ENGINE[1], while_text, problem_text, **kwargs)


class ErrTemplateIdNotUnique(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_ID_NOT_UNIQUE[0], ERR_TEMPLATE_ID_NOT_UNIQUE[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidKeypoint(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_KEYPOINT[0], ERR_TEMPLATE_INVALID_KEYPOINT[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidFeature(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_FEATURE[0], ERR_TEMPLATE_INVALID_FEATURE[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidConcurrencyConfig(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG[0], ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidSyntax(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_SYNTAX[0], ERR_TEMPLATE_INVALID_SYNTAX[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidFeatureClass(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_FEATURE_CLASS[0], ERR_TEMPLATE_INVALID_FEATURE_CLASS[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidMutator(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_MUTATOR[0], ERR_TEMPLATE_INVALID_MUTATOR[1], while_text, problem_text, **kwargs)


class ErrTemplateInvalidInterpretation(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_INTERPRETATION[0], ERR_TEMPLATE_INVALID_INTERPRETATION[1], while_text, problem_text, **kwargs)



"""
Template specifications have a lot of fields and properties.
Therefore, their proper validation is a serious task.
The main aim of this module is to build a schema for the entire template specification, so that it is
possible to automatically validate user imput against that schema, before continuing to process it.
"""


import abc
from typing import Dict

import cv2

from officialeye._internal.mutation.config import MutatorConfig


class Mutator(abc.ABC):

    def __init__(self, mutator_id: str, config_dict: Dict[str, any], /):
        super().__init__()

        self.mutator_id = mutator_id

        self._config = MutatorConfig(config_dict, mutator_id)

    def get_config(self) -> MutatorConfig:
        return self._config

    @abc.abstractmethod
    def mutate(self, img: cv2.Mat, /) -> cv2.Mat:
        raise NotImplementedError()



import strictyaml as yml

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidSyntax
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.template.schema.schema import generate_template_schema
from officialeye._internal.template.template import Template

_oe_template_schema = generate_template_schema()


def _print_error_message(err: yml.StrictYAMLError, template_path: str):

    get_logger().error("Error ", bold=True, nl=False)

    if err.context is not None:
        get_logger().error(err.context, prefix=False)
    else:
        get_logger().error("while parsing", prefix=False)

    if err.context_mark is not None and (
            err.problem is None
            or err.problem_mark is None
            or err.context_mark.name != err.problem_mark.name
            or err.context_mark.line != err.problem_mark.line
            or err.context_mark.column != err.problem_mark.column
    ):
        get_logger().error(str(err.context_mark).replace("<unicode string>", template_path))

    if err.problem is not None:
        get_logger().error("Problem", bold=True, nl=False)
        get_logger().error(f": {err.problem}", prefix=False)

    if err.problem_mark is not None:
        get_logger().error(str(err.problem_mark).replace("<unicode string>", template_path))


def load_template(context: Context, path: str) -> Template:
    """
    Loads a template from a file located at the specified path.

    Arguments:
        context: The global officialeye context.
        path: The path to the YAML template configuration file.

    Returns:
        Loaded template.

    Raises:
        OEError: In case there has been an error validating the correctness of the template.
    """

    global _oe_template_schema

    with open(path, "r") as fh:
        raw_data = fh.read()

    try:
        yaml_document = yml.load(raw_data, schema=_oe_template_schema)
    except yml.StrictYAMLError as err:
        _print_error_message(err, path)
        exit(4)
    except yml.YAMLError as err:
        raise ErrTemplateInvalidSyntax(
            f"while loading template configuration file at '{path}'.",
            "General parsing error. Check the syntax and the encoding of the file."
        ) from err

    data = yaml_document.data

    template = Template(context, data, path)

    get_logger().info("Loaded template: ", nl=False)
    get_logger().info(str(template), prefix=False, bold=True)

    return template



from typing import Callable

import click

from officialeye._internal.error.error import OEError
from officialeye.meta import OFFICIALEYE_CLI_LOGO


def _do_print_oe_error(output_func: Callable[[str], None], error: OEError, /):
    output_func(f"Error {error.code} in module {error.module}: {error.code_text}")
    output_func(f"Error occurred {error.while_text}")
    output_func(f"Problem: {error.problem_text}")

    causes = error.get_causes()
    external_causes = error.get_external_causes()

    if len(causes) + len(external_causes) > 0:
        output_func("The above error has been caused by the following error(s).")

    for cause in causes:
        _do_print_oe_error(output_func, cause)

    for external_cause in external_causes:
        output_func(str(external_cause))


class Logger:

    def __init__(self, /, *, debug_mode: bool = False, quiet_mode: bool = False, verbose_mode: bool = False, disable_logo: bool = False):

        self.debug_mode = debug_mode
        self.quiet_mode = quiet_mode
        self.verbose_mode = verbose_mode
        self.disable_logo = disable_logo

    def debug(self, msg, *args, prefix: bool = True, **kwargs):

        if not self.debug_mode:
            return

        if self.quiet_mode:
            return

        if prefix:
            click.secho("DEBUG", bold=True, bg="yellow", nl=False)
            click.echo(" ", nl=False)

        click.secho(msg, *args, **kwargs)

    def info(self, msg: any, *args, prefix: bool = True, **kwargs):

        if self.quiet_mode:
            return

        if prefix:
            click.secho("INFO", bold=True, bg="blue", nl=False)
            click.echo("  ", nl=False)

        click.secho(msg, *args, **kwargs)

    def warn(self, msg: any, *args, prefix: bool = True, **kwargs):

        if self.quiet_mode:
            return

        if prefix:
            click.secho("WARN", bold=True, bg="yellow", nl=False)
            click.echo("  ", nl=False)

        click.secho(msg, *args, **kwargs)

    def debug_verbose(self, msg: any, *args, prefix: bool = True, **kwargs):

        if not self.debug_mode:
            return

        if self.quiet_mode:
            return

        if not self.verbose_mode:
            return

        if prefix:
            click.secho("DEBUG", bold=True, bg="yellow", nl=False)
            click.echo(" ", nl=False)

        click.secho(msg, *args, **kwargs)

    def error(self, msg: any, *args, prefix: bool = True, **kwargs):

        if self.quiet_mode:
            return

        for line in msg.splitlines():
            if prefix:
                click.secho("ERROR", bold=True, bg="red", nl=False, err=True)
                click.echo(" ", nl=False, err=True)
            click.secho(line, *args, **kwargs, err=True)

    def logo(self):

        if self.quiet_mode:
            return

        if self.disable_logo:
            return

        click.secho(OFFICIALEYE_CLI_LOGO, fg="red")

    def error_oe_error(self, error: OEError, /):
        _do_print_oe_error(self.error, error)

    def debug_oe_error(self, error: OEError, /):
        _do_print_oe_error(self.debug, error)



import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidKeypoint
from officialeye._internal.template.region.region import TemplateRegion

_KEYPOINT_RECT_COLOR = (0, 0, 0xff)


class TemplateKeypoint(TemplateRegion):
    def __init__(self, context: Context, template_id: str, keypoint_dict: dict, /):
        super().__init__(context, template_id, keypoint_dict)

        self._matches_min = keypoint_dict["matches"]["min"]
        self._matches_max = keypoint_dict["matches"]["max"]

        if self._matches_max < self._matches_min:
            raise ErrTemplateInvalidKeypoint(
                f"while loading template keypoint '{self.region_id}'",
                f"the lower bound on the match count ({self._matches_min}) exceeds the upper bound ({self._matches_max})"
            )

        if self._matches_min < 0:
            raise ErrTemplateInvalidKeypoint(
                f"while loading template keypoint '{self.region_id}'",
                f"the lower bound on the match count ({self._matches_min}) cannot be negative"
            )

        assert 0 <= self._matches_min <= self._matches_max

    def get_matches_min(self) -> int:
        return self._matches_min

    def get_matches_max(self) -> int:
        return self._matches_max

    def visualize(self, img: cv2.Mat, /):
        return super()._visualize(img, rect_color=_KEYPOINT_RECT_COLOR)



__title__ = "officialeye"
__description__ = "An AI-powered generic document-analysis tool."
__url__ = "https://officialeye.zerobone.net"
__version__ = "1.1.5"
__author__ = "Alexander Mayorov"
__author_email__ = "zb@zerobone.net"
__license__ = "GPL-3.0"
__copyright__ = "Copyright Alexander Mayorov"



from typing import Dict

import cv2

from officialeye._internal.error.errors.template import ErrTemplateInvalidMutator
from officialeye._internal.mutation.mutator import Mutator


class RotateMutator(Mutator):
    """
    Mutator that rotates the image clockwise by the specified angle.
    """

    MUTATOR_ID = "rotate"

    def __init__(self, config: Dict[str, any], /):
        super().__init__(RotateMutator.MUTATOR_ID, config)

        def _angle_preprocessor(angle_text: str) -> int:
            angle = int(angle_text)

            if angle not in (0, 90, 180, 270):
                raise ErrTemplateInvalidMutator(
                    f"while loading mutator '{self.mutator_id}'.",
                    f"The 'angle' parameter must be a multiple of 90 degrees, got {angle}."
                )

            return angle

        self.get_config().set_value_preprocessor("angle", _angle_preprocessor)

        self._angle = self.get_config().get("angle")

    def mutate(self, img: cv2.Mat, /) -> cv2.Mat:

        if self._angle == 0:
            # we do not need to rotate the image at all
            return img

        if self._angle == 90:
            cv2_rotate_code = cv2.ROTATE_90_CLOCKWISE
        elif self._angle == 180:
            cv2_rotate_code = cv2.ROTATE_180
        elif self._angle == 270:
            cv2_rotate_code = cv2.ROTATE_90_COUNTERCLOCKWISE
        else:
            raise AssertionError()

        return cv2.rotate(img, cv2_rotate_code)



class DiffObjectException(Exception):
    """
    Base class for all `diffobject` exceptions.
    """

    def __init__(self, problem: str, /):
        super().__init__()

        self.problem = problem

    def __str__(self):
        return self.problem



import os

from officialeye._internal.error.errors.io import ErrIOInvalidPath
from officialeye._internal.logger.singleton import get_logger


def create_example_template_config_file(template_path: str, template_image: str, template_id: str, template_name: str, force_mode: bool, /):

    # validate the path first
    if os.path.isdir(template_path):
        raise ErrIOInvalidPath(
            f"while making sure that '{template_path}' references a file",
            "This path references a directory."
        )

    # determine source config value
    if os.path.isabs(template_image):
        # if the path to the specified template source file is absolute,
        # then we simply put the same absolute path in the template config
        template_source_config = template_image
    else:
        # otherwise, the path is relative to the working directory the program was started in
        # however, we are interested in the path to the template image relative to the template configuration file
        # therefore, we 'subtract' from the template_image path the template_path path
        template_source_config = os.path.relpath(template_image, os.path.dirname(template_path))

    if os.path.exists(template_path) and not force_mode:
        raise ErrIOInvalidPath(
            f"while making sure that path '{template_path}' exists",
            "There is already a file at this path. Use --force to suppress this error."
        )

    dir_name = os.path.dirname(template_path)
    if dir_name != "" and dir_name != "/" and not os.path.exists(dir_name):
        if force_mode:
            os.makedirs(dir_name, exist_ok=True)
        else:
            raise ErrIOInvalidPath(
                f"while making sure that path '{template_path}' exists",
                "Not all directories leading to the file exist. Use --force to create the missing directories."
            )

    template_yml = f'''# This is the ID of the template. It is needed to identify this template in the API and must be alphanumeric and unique.
id: "{template_id}"

# The name of the template, as should be displayed to the user
name: "{template_name}"

# This is the path to the example document you want to use as a base for the template.
# The path can be absolute or relative. If it is relative, then it will be resolved relative to the location of the current file.
source: "{template_source_config}"
# A mutator is a size-preserving image transforming function. Examples of mutators include noise reduction and color correction. 
mutators:
  # A list of mutators that should be applied to the above template source image before the unification phase
  source:
  # A list of mutators that should be applied to the input image before the unification phase
  target:

# A list of keypoints located in the template source image specified above.
# A keypoint is a rectangular region that should be present in all documents of this kind, and that
# should be used to find correspondences between the position of the given image and the positions in the template source image.
keypoints:
  # Example keypoint specification.
  # Remember to remove or customize this keypoint that is here only for demonstration purposes.
  example_keypoint_name:
    # Remember that the origin is located at the top left corner of the image. In other words,
    # the x-axis points to the right, and the y-axis points down (not up!).
    x: 10 # the x-coordinate of the top left corner of the rectangle (measured in pixels).
    y: 10 # the y-coordinate of the top left corner of the rectangle (measured in pixels).
    w: 100 # width of the rectangle (measured in pixels)
    h: 50 # height of the rectangle (measured in pixels)
    matches:
      # Minimum amount of matches that should be identified within this keypoint's region when analyzing an image.
      min: 0
      # Maximum amount of matches that should be identified within this keypoint's region when analyzing an image.
      max: 40

# Matching configuration
# Matching is the process of finding equal patterns in the given image and the template source image,
# thus establishing correspondences between positions.
# Although better matchings will of course contribute to the accuracy of the document analysis, it is
# completely fine for a matching algorithm to err in some cases.
matching:
  # Here you can specify the name of the matching engine that should be used to find correspondences between
  # positions of the given image and those of the template source image provided above.
  # Available engines: sift_flann
  engine: sift_flann
  # Engine-specific configuration
  config:
    # Configuration specific to the `sift_flann` matching engine.
    sift_flann:
      sensitivity: 0.7

# Supervision configuration
# Supervision is the process of unifying the given document image with the template source,
# based on the (partially unreliable) information received from the matching algorithm
supervision:
  # Here you can specify the name of the supervision engine that should be used to find correspondences between
  # positions of the given image and those of the template source image provided above.
  # Available engines: orthogonal_regression, least_squares_regression, combinatorial
  engine: combinatorial
  # Engine-specific configuration
  config:
    # Here you can specify engine-specific configurations
    combinatorial:
      # A coefficient between 0 and 1, representing the fraction of matches that should be chosen as the basis for inferring the way the document
      # is positioned in the image provided. The larger the coefficient, the more reliable the supervision result will be.
      # However, increasing this value will also increase the risk of wrongfully not detecting a document at all.
      # If your matching engine produces a lot of false positives, it is recommended to keep this value low, and vice versa.
      min_match_factor: 0.1
      # The combinatorial engine uses an affine linear transformation model to represent the way the document is located in the image provided.
      # It then chooses at least `min_match_factor` * (total amount of matches) matches that cause the data to fit the model, up to the error
      # specified here, measured in pixels along one of the axis. Thus, the lower this value is, the higher will be the quality of the
      # combinatorial supervision engine's output. On the other hand, a very small value will increase
      # the risk of wrongfully not detecting a document at all.
      # Recommended value: between 2 and 10 pixels
      max_transformation_error: 5

  # Since the supervision engine may produce multiple results.
  # This option allows you to specify the strategy that should be used to choose a final result.
  # Available strategies: first, best_mse, best_score
  # The `first` strategy simply returns the first result produced
  # The `random` strategy chooses a random result produced
  # The `best_mse` strategy finds the result which has the lowest mean squared error when translating template points to target points
  # The `best_score` strategy returns the result with the highest score (what `score` represents depends on the supervision engine)
  result: best_score

# A list of features located in the template source image specified above.
# A feature is a rectangular region containing any information of interest, such as text.
# In other words, the corresponding regions in the target image will be found during document analysis.
features:
  # Example feature specification.
  # Remember to remove or customize this feature that is here only for demonstration purposes.
  # Remember that the origin is located at the top left corner of the image. In other words,
  # the x-axis points to the right, and the y-axis points down (not up!).
  example_feature_name:
    # Remember that the origin is located at the top left corner of the image. In other words,
    # the x-axis points to the right, and the y-axis points down (not up!).
    x: 110 # the x-coordinate of the top left corner of the rectangle (measured in pixels)
    y: 10 # the y-coordinate of the top left corner of the rectangle (measured in pixels)
    w: 100 # width of the rectangle (measured in pixels)
    h: 10 # height of the rectangle (measured in pixels)
    class: example_feature_class # optional, the class of this feature (see below)

# A feature class can be used to group similar features together
# Feature classes are identified by a unique name, and can inherit other classes.
feature_classes:
  # For example, the following entry defines a feature class named example_feature_class
  example_feature_class:
    # An abstract class cannot be directly used by any concrete feature, but can be incomplete.
    # To use an abstract class, a non-abstract subclass has to be created.
    # A non-abstract class can be used by concrete features, but cannot be incomplete.
    abstract: no # Optional. By default, feature classes are not abstract.
    # A mutator is a method of transforming an image into another image.
    # Examples of mutators include image denoising, grayscale coloring, etc.
    # You can even define and use your own mutators.
    # In this example, we use a mutator that reduces image noise.
    mutators: # A sequence of mutators that should be applied to the region of a feature, before it is processed further.
      - id: non_local_means_denoising # Name of the mutator
        config: # Optional mutator-specific configuration
          colored: yes
    interpretation:
      # An interpretation method defines the way in which the mutated feature location should be processed further
      # For example, the ocr_tesseract method will apply the Tesseract OCR to the image.
      method: ocr_tesseract
      # Intepretation-engine-specific configuration values
      config:
        config: --dpi 10000 --oem 3 --psm 6
        lang: eng
    '''

    with open(template_path, "w") as fh:
        fh.write(template_yml)

    get_logger().info(f"Initialized template configuration file at '{template_path}'.")



from officialeye._internal.error.codes import (
    ERR_IO_INVALID_IMAGE,
    ERR_IO_INVALID_PATH,
    ERR_IO_INVALID_SUPERVISION_ENGINE,
    ERR_IO_OPERATION_NOT_SUPPORTED_BY_DRIVER,
)
from officialeye._internal.error.error import OEError
from officialeye._internal.error.modules import ERR_MODULE_IO


class ErrIO(OEError):

    def __init__(self, code: int, code_text: str, while_text: str, problem_text: str, /, *, is_regular: bool = False, **kwargs):
        super().__init__(ERR_MODULE_IO, code, code_text, while_text, problem_text, is_regular=is_regular)


class ErrIOInvalidSupervisionEngine(ErrIO):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_IO_INVALID_SUPERVISION_ENGINE[0], ERR_IO_INVALID_SUPERVISION_ENGINE[1], while_text, problem_text, **kwargs)


class ErrIOOperationNotSupportedByDriver(ErrIO):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_IO_OPERATION_NOT_SUPPORTED_BY_DRIVER[0], ERR_IO_OPERATION_NOT_SUPPORTED_BY_DRIVER[1], while_text, problem_text, **kwargs)


class ErrIOInvalidPath(ErrIO):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_IO_INVALID_PATH[0], ERR_IO_INVALID_PATH[1], while_text, problem_text, **kwargs)


class ErrIOInvalidImage(ErrIO):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_IO_INVALID_IMAGE[0], ERR_IO_INVALID_IMAGE[1], while_text, problem_text, **kwargs)



from officialeye._internal.error.codes import ERR_MATCHING_INVALID_ENGINE_CONFIG, ERR_MATCHING_MATCH_COUNT_OUT_OF_BOUNDS
from officialeye._internal.error.error import OEError
from officialeye._internal.error.modules import ERR_MODULE_MATCHING


class ErrMatching(OEError):

    def __init__(self, code: int, code_text: str, while_text: str, problem_text: str, /, *, is_regular: bool, **kwargs):
        super().__init__(ERR_MODULE_MATCHING, code, code_text, while_text, problem_text, is_regular=is_regular)


class ErrMatchingMatchCountOutOfBounds(ErrMatching):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_MATCHING_MATCH_COUNT_OUT_OF_BOUNDS[0], ERR_MATCHING_MATCH_COUNT_OUT_OF_BOUNDS[1], while_text, problem_text, is_regular=True, **kwargs)


class ErrMatchingInvalidEngineConfig(ErrMatching):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_MATCHING_INVALID_ENGINE_CONFIG[0], ERR_MATCHING_INVALID_ENGINE_CONFIG[1], while_text, problem_text, is_regular=False, **kwargs)



from typing import Dict, List

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.matching import ErrMatchingMatchCountOutOfBounds
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.match import Match


class MatchingResult:

    def __init__(self, context: Context, template_id: str, /):
        self._context = context
        self._template_id = template_id

        # keys: keypoint ids
        # values: matches with this keypoint
        self._matches_dict: Dict[str, List[Match]] = {}

        for keypoint in self.get_template().keypoints():
            self._matches_dict[keypoint.region_id] = []

    def remove_all_matches(self):
        self._matches_dict = {}

    def add_match(self, match: Match, /):
        assert match.keypoint_id in self._matches_dict
        self._matches_dict[match.keypoint_id].append(match)

    def get_matches(self):
        for keypoint_id in self._matches_dict:
            for match in self._matches_dict[keypoint_id]:
                yield match

    def get_total_match_count(self) -> int:
        match_count = 0
        for keypoint_id in self._matches_dict:
            match_count += len(self._matches_dict[keypoint_id])
        return match_count

    def get_keypoint_ids(self):
        for keypoint_id in self._matches_dict:
            yield keypoint_id

    def matches_for_keypoint(self, keypoint_id: str, /):
        for match in self._matches_dict[keypoint_id]:
            yield match

    def get_template(self):
        return self._context.get_template(self._template_id)

    def validate(self):

        get_logger().debug("Validating the keypoint matching result.")

        assert len(self._matches_dict) > 0

        total_match_count = 0

        # verify that for every keypoint, it has been matched a number of times that is in the desired bounds
        for keypoint_id in self._matches_dict:
            keypoint = self.get_template().get_keypoint(keypoint_id)

            keypoint_matches_min = keypoint.get_matches_min()
            keypoint_matches_max = keypoint.get_matches_max()

            keypoint_matches_count = len(self._matches_dict[keypoint_id])

            if keypoint_matches_count < keypoint_matches_min:
                raise ErrMatchingMatchCountOutOfBounds(
                    f"while checking that keypoint '{keypoint_id}' of template '{self._template_id}' "
                    f"has been matched a sufficient number of times",
                    f"Expected at least {keypoint_matches_min} matches, got {keypoint_matches_count}"
                )

            if keypoint_matches_count > keypoint_matches_max:
                get_logger().debug(
                    f"Keypoint '{keypoint_id}' of template '{self._template_id}' has too many matches "
                    f"(matches: {keypoint_matches_count} max: {keypoint_matches_max}). Cherry-picking the best matches.")
                # cherry-pick the best matches
                self._matches_dict[keypoint_id] = sorted(self._matches_dict[keypoint_id])[:keypoint_matches_max]
                keypoint_matches_count = keypoint_matches_max

            get_logger().debug(f"Keypoint '{keypoint_id}' of template '{self._template_id}' has been matched {keypoint_matches_count} times "
                               f"(min: {keypoint_matches_min} max: {keypoint_matches_max}).")

            total_match_count += keypoint_matches_count

        assert total_match_count >= 0
        if total_match_count == 0:
            raise ErrMatchingMatchCountOutOfBounds(
                f"while checking that there has been at least one match for template '{self._template_id}'.",
                "There have been no matches."
            )

    def debug_print(self):
        get_logger().debug(f"Found {self.get_total_match_count()} matched points!")

        get_logger().debug_verbose("Listing matched points:")
        for match in self.get_matches():
            get_logger().debug_verbose(f"> {match}")



import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.error import OEError
from officialeye._internal.io.driver import IODriver
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.template.template import Template


class TestIODriver(IODriver):

    def __init__(self, context: Context, /):
        super().__init__(context)

        self.visualize_features: bool = True

    def handle_supervision_result(self, target: cv2.Mat, result: SupervisionResult, /):

        assert result is not None

        template = self._context.get_template(result.template_id)

        application_image = template.load_source_image()

        # extract the features from the target image
        for feature in template.features():
            feature_img = result.get_feature_warped_region(target, feature)

            feature_img_mutated = feature.apply_mutators_to_image(feature_img)

            if feature_img.shape == feature_img_mutated.shape:
                # mutators didn't change the shape of the image
                feature.insert_into_image(application_image, feature_img_mutated)
            else:
                # some mutator has altered the shape of the feature image.
                # this means that we can no longer safely insert the mutated feature into the visualization.
                # therefore, we have to fall back to inserting the feature image unmutated
                get_logger().warn(f"Could not visualize the '{feature.region_id}' feature of the '{feature.get_template().template_id}' template, "
                                  f"because one of the mutators (corresponding to this feature) did not preserve the shape of the image. "
                                  f"Falling back to the non-mutated version of the feature image.")
                feature.insert_into_image(application_image, feature_img)

        if self.visualize_features:
            # visualize features on the image
            for feature in template.features():
                application_image = feature.visualize(application_image)

        self._context.export_primary_image(application_image, file_name="supervision_result.png")

    def handle_show_result(self, template: Template, img: cv2.Mat, /):
        self._context.export_primary_image(img, file_name=f"{template.template_id}.png")

    def handle_error(self, error: OEError, /):
        get_logger().error_oe_error(error)



import os
from typing import Dict

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidInterpretation
from officialeye._internal.interpretation.method import InterpretationMethod
from officialeye._internal.interpretation.serializable import Serializable


class FileMethod(InterpretationMethod):

    METHOD_ID = "file"

    def __init__(self, context: Context, config_dict: Dict[str, any]):
        super().__init__(context, FileMethod.METHOD_ID, config_dict)

        self._path = self.get_config().get("path")

    def interpret(self, feature_img: cv2.Mat, template_id: str, feature_id: str, /) -> Serializable:

        feature = self._context.get_template(template_id).get_feature(feature_id)

        feature_class_generator = feature.get_feature_class().get_features()

        # check if the generator generates at least two elements
        if sum(1 for _ in zip(range(2), feature_class_generator, strict=False)) == 2:
            raise ErrTemplateInvalidInterpretation(
                "while applying the '{FileMethod.METHOD_ID}' interpretation method.",
                "This method cannot be applied if there are at least two features inheriting the feature class defining this method."
            )

        os.makedirs(os.path.dirname(self._path), exist_ok=True)

        cv2.imwrite(self._path, feature_img)

        return None



from typing import Dict

from officialeye._internal.config.config import Config
from officialeye._internal.error.errors.template import ErrTemplateInvalidMutator


class MutatorConfig(Config):

    def __init__(self, config_dict: Dict[str, any], mutator_id: str, /):

        super().__init__(config_dict)

        self._mutator_id = mutator_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrTemplateInvalidMutator(
            f"while reading configuration of the '{self._mutator_id}' mutator.",
            f"Could not find a value for key '{key}'."
        )



import numpy as np

from officialeye._internal.context.context import Context
from officialeye._internal.template.region.keypoint import TemplateKeypoint


class Match:

    def __init__(self, context: Context, template_id: str, keypoint_region_id: str,
                 region_point: np.ndarray, target_point: np.ndarray, /, *, score: float = 0.0):

        self._context = context

        self.template_id = template_id
        self.keypoint_id = keypoint_region_id

        assert region_point.shape[0] == 2
        assert target_point.shape[0] == 2

        self._region_point = region_point
        self._target_point = target_point

        self._score = score

    def set_score(self, new_score: float, /):
        self._score = new_score

    def get_score(self) -> float:
        return self._score

    def get_template_point(self) -> np.ndarray:
        return self._region_point.copy()

    def get_template(self):
        return self._context.get_template(self.template_id)

    def get_keypoint(self) -> TemplateKeypoint:
        return self.get_template().get_keypoint(self.keypoint_id)

    def get_original_template_point(self) -> np.ndarray:
        """Returns the coordinates of the point lying in the keypoint, in the coordinate system of the underlying template."""
        return self._region_point + self.get_keypoint().get_top_left_vec()

    def get_target_point(self) -> np.ndarray:
        return self._target_point.copy()

    def __lt__(self, other) -> bool:
        assert isinstance(other, Match)
        return self.get_score() < other.get_score()

    def __eq__(self, o):
        if not isinstance(o, Match):
            return False
        if self.template_id != o.template_id:
            return False
        if self.keypoint_id != o.keypoint_id:
            return False
        return (np.array_equal(self._region_point, o._region_point)
                and np.array_equal(self._target_point, o._target_point))

    def __ne__(self, __value):
        return not self == __value

    def __hash__(self):
        return hash((self.template_id, self.keypoint_id, np.dot(self._region_point, self._target_point)))

    def __str__(self) -> str:
        return "%s_%s: (%4d, %4d) <-> (%4d, %4d)" % (self.template_id, self.keypoint_id,
                                                     int(self._region_point[0]), int(self._region_point[1]),
                                                     int(self._target_point[0]), int(self._target_point[1]))

    def get_debug_identifier(self) -> str:
        return "%s_%s_%04d_%04d_%04d_%04d" % (self.template_id, self.keypoint_id,
                                              int(self._region_point[0]), int(self._region_point[1]),
                                              int(self._target_point[0]), int(self._target_point[1]))



from typing import Union

import strictyaml as yml

from officialeye._internal.diffobject.difference_modes import DIFF_MODE_ADD, DIFF_MODE_OVERRIDE, DIFF_MODE_REMOVE
from officialeye._internal.diffobject.exception import DiffObjectException
from officialeye._internal.diffobject.specification_entry import DiffObjectSpecificationEntry


class ObjectSpecificationEntry(DiffObjectSpecificationEntry):

    def __init__(self, validator: yml.Validator, /):
        super().__init__(validator)

    def apply_diff(self, current_value: Union[dict, None], diff_value: dict, diff_mode: str) -> dict:
        assert current_value is None or isinstance(current_value, dict)
        assert isinstance(diff_value, dict)

        if diff_mode == DIFF_MODE_OVERRIDE:
            return diff_value

        if current_value is None:
            raise DiffObjectException(f"Could not apply difference mode '{diff_mode}' because the previous value is not available.")

        if diff_mode == DIFF_MODE_ADD:
            # merge the two dictionaries while prioritizing the `diff_value` entries
            return {**current_value, **diff_value}

        if diff_mode == DIFF_MODE_REMOVE:
            new_value = {}
            for key in current_value:
                if key in diff_value:
                    continue
                new_value[key] = current_value[key]
            return new_value

        raise DiffObjectException(f"The list type is incompatible with difference mode specification '{diff_mode}'.")



import abc
from typing import Dict, Tuple

import cv2
import numpy as np

from officialeye._internal.context.context import Context

_LABEL_COLOR_DEFAULT = (0, 0, 0xff)
_VISUALIZATION_SCALE_COEFF = 1.0 / 1400.0


class TemplateRegion(abc.ABC):

    def __init__(self, context: Context, template_id: str, region_dict: Dict[str, any], /):
        self._context = context
        self._template_id = template_id

        self.region_id = str(region_dict["id"])
        self.x = int(region_dict["x"])
        self.y = int(region_dict["y"])
        self.w = int(region_dict["w"])
        self.h = int(region_dict["h"])

    def get_template(self):
        return self._context.get_template(self._template_id)

    @abc.abstractmethod
    def visualize(self, img: cv2.Mat) -> cv2.Mat:
        raise NotImplementedError()

    def get_top_left_vec(self) -> np.ndarray:
        return np.array([self.x, self.y])

    def get_top_right_vec(self) -> np.ndarray:
        return np.array([self.x + self.w, self.y])

    def get_bottom_left_vec(self) -> np.ndarray:
        return np.array([self.x, self.y + self.h])

    def get_bottom_right_vec(self) -> np.ndarray:
        return np.array([self.x + self.w, self.y + self.h])

    def _visualize(self, img: cv2.Mat, /, *,
                   rect_color: Tuple[int, int, int], label_color=_LABEL_COLOR_DEFAULT) -> cv2.Mat:
        img = cv2.rectangle(img, (self.x, self.y), (self.x + self.w, self.y + self.h), rect_color, 4)
        label_origin = (
            self.x + int(10 * img.shape[0] * _VISUALIZATION_SCALE_COEFF),
            self.y + int(30 * img.shape[0] * _VISUALIZATION_SCALE_COEFF)
        )
        font_scale = img.shape[0] * _VISUALIZATION_SCALE_COEFF
        img = cv2.putText(
            img,
            self.region_id,
            label_origin,
            cv2.FONT_HERSHEY_SIMPLEX,
            font_scale,
            label_color,
            int(2 * img.shape[0] * _VISUALIZATION_SCALE_COEFF),
            cv2.LINE_AA
        )
        return img

    def to_image(self):
        img = self.get_template().load_source_image()
        return img[self.y:self.y + self.h, self.x:self.x + self.w]

    def insert_into_image(self, target: np.ndarray, transformed_version: np.ndarray = None):

        assert target.shape[0] == self.get_template().height
        assert target.shape[1] == self.get_template().width

        if transformed_version is None:
            transformed_version = self.to_image()

        target[self.y: self.y + self.h, self.x: self.x + self.w] = transformed_version



import abc
from typing import Dict

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.interpretation.config import InterpretationMethodConfig
from officialeye._internal.interpretation.serializable import Serializable


class InterpretationMethod(abc.ABC):

    def __init__(self, context: Context, method_id: str, config_dict: Dict[str, any], /):
        super().__init__()

        self._context = context
        self.method_id = method_id

        self._config = InterpretationMethodConfig(config_dict, method_id)

    def get_config(self) -> InterpretationMethodConfig:
        return self._config

    @abc.abstractmethod
    def interpret(self, feature_img: cv2.Mat, template_id: str, feature_id: str, /) -> Serializable:
        raise NotImplementedError()



from typing import Dict

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidInterpretation
from officialeye._internal.interpretation.method import InterpretationMethod
from officialeye._internal.interpretation.methods.file import FileMethod
from officialeye._internal.interpretation.methods.file_temp import FileTempMethod
from officialeye._internal.interpretation.methods.ocr_tesseract import TesseractMethod


def load_interpretation_method(context: Context, method_id: str, config_dict: Dict[str, any], /) -> InterpretationMethod:

    if method_id == TesseractMethod.METHOD_ID:
        return TesseractMethod(context, config_dict)

    if method_id == FileTempMethod.METHOD_ID:
        return FileTempMethod(context, config_dict)

    if method_id == FileMethod.METHOD_ID:
        return FileMethod(context, config_dict)

    raise ErrTemplateInvalidInterpretation(
        f"while loading interpretation method '{method_id}'.",
        "Unknown interpretation method id."
    )



import sys
from typing import Dict, Set

import cv2
import numpy as np

from officialeye._internal.matching.match import Match
from officialeye._internal.matching.result import MatchingResult


class SupervisionResult:

    def __init__(self, template_id: str, kmr: MatchingResult,
                 delta: np.ndarray, delta_prime: np.ndarray, transformation_matrix: np.ndarray, /):

        self.template_id = template_id
        self._kmr = kmr

        assert delta.shape == (2,)
        assert delta_prime.shape == (2,)
        assert transformation_matrix.shape == (2, 2)

        # offset in the template's coordinates
        self._delta = delta
        # offset in the target image's coordinates
        self._delta_prime = delta_prime
        # self.dpo = delta_prime.copy()

        self._transformation_matrix = transformation_matrix

        # keys: matches
        # values: weights assigned by the supervision engine to each match (assigning is optional)
        # the higher the weight, the more we trust the correctness of the match and the greater its individual impact should be.
        # by default, the weight is 1.
        self._match_weights: Dict[Match, float] = {}

        # an optional value the supervision engine can set, representing how confident the engine is that the result is of high quality
        self._score = 0.0

    def get_score(self) -> float:
        assert self._score >= 0.0
        return self._score

    def set_score(self, new_score: float, /):
        assert new_score >= 0
        assert isinstance(new_score, float)
        self._score = new_score

    def get_match_weight(self, match: Match, /) -> float:
        if match in self._match_weights:
            return self._match_weights[match]
        return 1.0

    def set_match_weight(self, match: Match, weight: float, /):
        self._match_weights[match] = weight

    def template_point_to_target_point(self, template_point: np.ndarray, /) -> np.ndarray:
        assert template_point.shape == (2,)
        assert self._delta.shape == (2,)
        assert self._delta_prime.shape == (2,)
        return self._transformation_matrix @ (template_point - self._delta) + self._delta_prime

    def get_feature_warped_region(self, target: cv2.Mat, feature) -> cv2.Mat:

        feature_tl = feature.get_top_left_vec()
        feature_tr = feature.get_top_right_vec()
        feature_bl = feature.get_bottom_left_vec()
        feature_br = feature.get_bottom_right_vec()

        target_tl = self.template_point_to_target_point(feature_tl)
        target_tr = self.template_point_to_target_point(feature_tr)
        target_bl = self.template_point_to_target_point(feature_bl)
        target_br = self.template_point_to_target_point(feature_br)

        dest_tl = np.array([0, 0], dtype=np.float64)
        dest_tr = np.array([feature.w, 0], dtype=np.float64)
        dest_br = np.array([feature.w, feature.h], dtype=np.float64)
        dest_bl = np.array([0, feature.h], dtype=np.float64)

        source_points = [target_tl, target_tr, target_br, target_bl]
        destination_points = [dest_tl, dest_tr, dest_br, dest_bl]

        homography = cv2.getPerspectiveTransform(np.float32(source_points), np.float32(destination_points))

        return cv2.warpPerspective(
            target,
            np.float32(homography),
            (feature.w, feature.h),
            flags=cv2.INTER_LINEAR
        )

    def get_relevant_keypoint_ids(self) -> Set[str]:
        rk = set()
        for match in self._kmr.get_matches():
            rk.add(match.get_keypoint().region_id)
        assert len(rk) > 0
        return rk

    def get_keypoint_matching_result(self) -> MatchingResult:
        return self._kmr

    def get_weighted_mse(self, /) -> float:
        error = 0.0
        singificant_match_count = 0
        for match in self._kmr.get_matches():

            match_weight = self.get_match_weight(match)

            if match_weight < sys.float_info.epsilon:
                continue

            singificant_match_count += 1

            s = match.get_original_template_point()
            # calculate prediction
            p = self.template_point_to_target_point(s)
            # calculate destination
            d = match.get_target_point()
            current_error = p - d
            current_error_value = np.dot(current_error, current_error)
            error += current_error_value * match_weight

        return error / singificant_match_count



"""
Module for abstracting out the ability to inject custom configurations specified using dictionaries.
The goal of this module is to provide a nice API for validated user-specified configurations
and safely retrieving information from there.
"""



from typing import Dict

import cv2
from pytesseract import pytesseract

from officialeye._internal.context.context import Context
from officialeye._internal.interpretation.method import InterpretationMethod
from officialeye._internal.interpretation.serializable import Serializable


class TesseractMethod(InterpretationMethod):

    METHOD_ID = "ocr_tesseract"

    def __init__(self, context: Context, config_dict: Dict[str, any]):
        super().__init__(context, TesseractMethod.METHOD_ID, config_dict)

        self._tesseract_lang = self.get_config().get("lang", default="eng")
        self._tesseract_config = self.get_config().get("config", default="")

    def interpret(self, feature_img: cv2.Mat, template_id: str, feature_id: str, /) -> Serializable:
        return pytesseract.image_to_string(feature_img, lang=self._tesseract_lang, config=self._tesseract_config).strip()



import random
from typing import Dict, Generator

import numpy as np
import z3

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.supervision import ErrSupervisionInvalidEngineConfig
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.match import Match
from officialeye._internal.matching.result import MatchingResult
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.supervision.supervisor import Supervisor


class CombinatorialSupervisor(Supervisor):

    ENGINE_ID = "combinatorial"

    def __init__(self, context: Context, template_id: str, kmr: MatchingResult, /):
        super().__init__(context, CombinatorialSupervisor.ENGINE_ID, template_id, kmr)

        # setup configuration
        def _min_match_factor_preprocessor(v: str) -> float:

            v = float(v)

            if v > 1.0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.ENGINE_ID}' supervisor",
                    f"The `min_match_factor` value ({v}) cannot exceed 1.0."
                )

            if v < 0.0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.ENGINE_ID}' supervisor",
                    f"The `min_match_factor` value ({v}) cannot be negative."
                )

            return v

        self.get_config().set_value_preprocessor("min_match_factor", _min_match_factor_preprocessor)

        def _max_transformation_error_preprocessor(v: str) -> int:

            v = int(v)

            if v < 0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.ENGINE_ID}' supervisor.",
                    f"The `max_transformation_error` value ({v}) cannot be negative."
                )

            if v > 5000:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.ENGINE_ID}' supervisor.",
                    f"The `max_transformation_error` value ({v}) is too high."
                )

            return v

        self.get_config().set_value_preprocessor("max_transformation_error", _max_transformation_error_preprocessor)

        def _z3_timeout_preprocessor(v: str) -> int:

            v = int(v)

            if v < 1:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.ENGINE_ID}' supervisor.",
                    f"The `z3_timeout` value ({v}) cannot be negative or zero."
                )

            return v

        self.get_config().set_value_preprocessor("z3_timeout", _z3_timeout_preprocessor)

        # initialize all engine-specific values
        self._z3_context = z3.Context()

        # create variables for components of the translation matrix
        self._transformation_matrix = np.array([
            [z3.Real("a", ctx=self._z3_context), z3.Real("b", ctx=self._z3_context)],
            [z3.Real("c", ctx=self._z3_context), z3.Real("d", ctx=self._z3_context)]
        ], dtype=z3.AstRef)

        # keys: matches (instances of Match)
        # values: z3 integer variables representing the errors for each match,
        # i.e., how consistent the match is with the affine transformation model
        self._match_weight: Dict[Match, z3.ArithRef] = {}

        for match in self._kmr.get_matches():
            self._match_weight[match] = z3.Real(f"w_{match.get_debug_identifier()}", ctx=self._z3_context)

        _config_min_match_factor = self.get_config().get("min_match_factor", default=0.1)
        get_logger().debug(f"Min match factor: {_config_min_match_factor}")

        self._minimum_weight_to_enforce = self._kmr.get_total_match_count() * _config_min_match_factor
        self._max_transformation_error = self.get_config().get("max_transformation_error")

    def _get_consistency_check(self, match: Match, delta: np.ndarray, delta_prime: np.ndarray, /) -> z3.AstRef:
        """
        Generates a z3 formula asserting the consistency of the match with the affine linear transformation model.
        Consistency does not mean ideal matching of coordinates; rather, the template position with the affine
        transformation applied to it, must roughly be equal the target position for consistency to hold
        In other words, targetpoint = M * (templatepoint - offset), where offset is a vector and M is a 2x2 matrix
        """

        template_point = match.get_original_template_point()

        assert delta.shape == (2,)
        assert delta_prime.shape == (2,)
        assert template_point.shape == (2,)

        translated_template_point = self._transformation_matrix @ (template_point - delta) + delta_prime
        translated_template_point_x, translated_template_point_y = translated_template_point

        target_point_x, target_point_y = match.get_target_point()

        return z3.And(
            translated_template_point_x - target_point_x <= self._max_transformation_error,
            target_point_x - translated_template_point_x <= self._max_transformation_error,
            translated_template_point_y - target_point_y <= self._max_transformation_error,
            target_point_y - translated_template_point_y <= self._max_transformation_error,
        )

    def _run(self) -> Generator[SupervisionResult, None, None]:

        weights_lower_bounds = z3.And(*(self._match_weight[match] >= 0 for match in self._kmr.get_matches()), self._z3_context)
        weights_upper_bounds = z3.And(*(self._match_weight[match] <= 1 for match in self._kmr.get_matches()), self._z3_context)

        total_weight = z3.Sum(*(self._match_weight[match] for match in self._kmr.get_matches()))

        solver = z3.Optimize(ctx=self._z3_context)
        solver.set("timeout", self.get_config().get("z3_timeout", default=2500))

        solver.add(weights_lower_bounds)
        solver.add(weights_upper_bounds)
        solver.add(total_weight >= self._minimum_weight_to_enforce)

        solver.maximize(total_weight)

        for keypoint_id in self._kmr.get_keypoint_ids():
            keypoint_matches = list(self._kmr.matches_for_keypoint(keypoint_id))

            if len(keypoint_matches) == 0:
                continue

            # TODO: think whether this is a good algorithm design decision, and improve it if not
            anchor_match = keypoint_matches[random.randint(0, len(keypoint_matches) - 1)]

            delta = anchor_match.get_original_template_point()
            delta_prime = anchor_match.get_target_point()

            solver.push()

            for match in self._kmr.get_matches():
                solver.add(z3.Implies(
                    self._match_weight[match] > 0,
                    # consistency check
                    self._get_consistency_check(match, delta, delta_prime),
                    ctx=self._z3_context
                ))

            result = solver.check()

            if result == z3.unsat:
                get_logger().warn("Could not satisfy the imposed constraints.", fg="red")
                solver.pop()
                continue

            if result == z3.unknown:
                get_logger().warn("Could not decide the satifiability of the imposed constraints.", fg="red")
                solver.pop()
                continue

            assert result == z3.sat

            model = solver.model()
            model_evaluator = np.vectorize(lambda var: model.eval(var, model_completion=True).as_fraction())  # noqa: B023

            # extract total weight and maximization target from the model
            model_total_weight = float(model_evaluator(total_weight))

            # extract transformation matrix from model
            transformation_matrix = model_evaluator(self._transformation_matrix)

            _result = SupervisionResult(self.template_id, self._kmr, delta, delta_prime, transformation_matrix)
            # add fixed constant to make sure that the score value is always non-negative
            _result.set_score(model_total_weight)

            for match in self._kmr.get_matches():
                match_weight = model_evaluator(self._match_weight[match])
                _result.set_match_weight(match, match_weight)

            get_logger().debug(f"Error: {_result.get_weighted_mse()} Total weight and score: {model_total_weight}")

            yield _result

            solver.pop()



import abc
import random
from abc import ABC
from typing import Generator, Union

from officialeye._internal.context.context import Context
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.matching.result import MatchingResult
from officialeye._internal.supervision.result import SupervisionResult
from officialeye._internal.supervision.supervisor_config import SupervisorConfig

_SUPERVISION_RESULT_FIRST = "first"
_SUPERVISION_RESULT_RANDOM = "random"
_SUPERVISION_RESULT_BEST_MSE = "best_mse"
_SUPERVISION_RESULT_BEST_SCORE = "best_score"


class Supervisor(ABC):

    def __init__(self, context: Context, engine_id: str, template_id: str, kmr: MatchingResult, /):
        super().__init__()

        self._context = context
        self._engine_id = engine_id

        self.template_id = template_id
        self._kmr = kmr

        get_logger().debug(f"Total match count: {self._kmr.get_total_match_count()}")

        # initialize configuration manager
        supervision_config = self.get_template().get_supervision_config()

        assert isinstance(supervision_config, dict)

        if self._engine_id in supervision_config:
            config_dict = supervision_config[self._engine_id]
        else:
            get_logger().warn(f"Could not find any configuration entries for the '{self._engine_id}' supervision engine.")
            config_dict = {}

        self._config = SupervisorConfig(config_dict, self._engine_id)

    def get_template(self):
        return self._context.get_template(self.template_id)

    @abc.abstractmethod
    def _run(self) -> Generator[SupervisionResult, None, None]:
        raise NotImplementedError()

    def _run_first(self) -> Union[SupervisionResult, None]:
        results_generator = self._run()
        return next(results_generator, None)

    def _run_random(self) -> Union[SupervisionResult, None]:
        results = list(self._run())
        return None if len(results) == 0 else results[random.randint(0, len(results) - 1)]

    def _run_best_mse(self) -> Union[SupervisionResult, None]:

        results = list(self._run())

        if len(results) == 0:
            return None

        best_result = results[0]
        best_result_mse = best_result.get_weighted_mse()

        for result_id, result in enumerate(results):
            result_mse = result.get_weighted_mse()

            get_logger().debug_verbose(f"Result #{result_id + 1} has MSE {result_mse}")

            if result_mse < best_result_mse:
                best_result_mse = result_mse
                best_result = result

        get_logger().debug(f"Best result has MSE {best_result_mse}")

        return best_result

    def _run_best_score(self) -> Union[SupervisionResult, None]:

        results = list(self._run())

        if len(results) == 0:
            return None

        best_result = results[0]
        best_result_score = best_result.get_score()
        best_result_mse = best_result.get_weighted_mse()

        for result_id, result in enumerate(results):
            result_score = result.get_score()

            get_logger().debug_verbose(f"Result #{result_id + 1} has score {result_score}")

            if result_score > best_result_score:
                best_result_score = result_score
                best_result_mse = result.get_weighted_mse()
                best_result = result
            elif result_score == best_result_score:
                current_result_mse = result.get_weighted_mse()
                if current_result_mse < best_result_mse:
                    best_result_mse = current_result_mse
                    best_result = result

        get_logger().debug(f"Best result has score {best_result_score} and MSE {best_result_mse}")

        return best_result

    def get_config(self) -> SupervisorConfig:
        return self._config

    def run(self) -> Union[SupervisionResult, None]:

        supervision_result_choice_engine = self.get_template().get_supervision_result()

        get_logger().debug(f"Applying '{supervision_result_choice_engine}' supervision result choice engine.")

        if supervision_result_choice_engine == _SUPERVISION_RESULT_FIRST:
            return self._run_first()

        if supervision_result_choice_engine == _SUPERVISION_RESULT_RANDOM:
            return self._run_random()

        if supervision_result_choice_engine == _SUPERVISION_RESULT_BEST_MSE:
            return self._run_best_mse()

        if supervision_result_choice_engine == _SUPERVISION_RESULT_BEST_SCORE:
            return self._run_best_score()

        raise AssertionError("Invalid supervision result")



from typing import Dict, Union

import cv2

from officialeye._internal.context.context import Context
from officialeye._internal.error.errors.template import ErrTemplateInvalidFeature
from officialeye._internal.interpretation.loader import load_interpretation_method
from officialeye._internal.logger.singleton import get_logger
from officialeye._internal.mutation.loader import load_mutator_from_dict
from officialeye._internal.template.feature_class.feature_class import FeatureClass
from officialeye._internal.template.feature_class.manager import FeatureClassManager
from officialeye._internal.template.region.region import TemplateRegion

_FEATURE_RECT_COLOR = (0, 0xff, 0)


class TemplateFeature(TemplateRegion):

    def __init__(self, context: Context, template_id: str, feature_dict: Dict[str, any], /):
        super().__init__(context, template_id, feature_dict)

        if "class" in feature_dict:
            self._class_id = feature_dict["class"]
            assert isinstance(self._class_id, str)
        else:
            self._class_id = None

    def visualize(self, img: cv2.Mat, /):
        return super()._visualize(img, rect_color=_FEATURE_RECT_COLOR)

    def validate_feature_class(self):

        if self._class_id is None:
            return

        feature_classes: FeatureClassManager = self.get_template().get_feature_classes()

        if not feature_classes.contains_class(self._class_id):
            raise ErrTemplateInvalidFeature(
                f"while loading class for feature '{self.region_id}' in template '{self.get_template().template_id}'.",
                f"Specified feature class '{self._class_id}' is not defined."
            )

        feature_class = feature_classes.get_class(self._class_id)

        if feature_class.is_abstract():
            raise ErrTemplateInvalidFeature(
                f"while loading class for feature '{self.region_id}' in template '{self.get_template().template_id}'.",
                f"Cannot instantiate an abstract feature class '{self._class_id}'."
            )

    def get_feature_class(self) -> Union[FeatureClass, None]:
        """ Returns class of feature, or None if the feature does not have a class. """

        if self._class_id is None:
            return None

        feature_classes: FeatureClassManager = self.get_template().get_feature_classes()

        assert feature_classes.contains_class(self._class_id)

        feature_class = feature_classes.get_class(self._class_id)

        assert not feature_class.is_abstract()

        return feature_class

    def apply_mutators_to_image(self, img: cv2.Mat, /) -> cv2.Mat:
        """
        Takes an image and applies the mutators defined in the corresponding feature class.

        Arguments:
            img: The image that should be transformed.

        Returns:
            The resulting image.
        """

        feature_class = self.get_feature_class()

        if feature_class is None:
            return img

        mutators = feature_class.get_data()["mutators"]

        assert isinstance(mutators, list)

        for mutator_dict in mutators:
            mutator = load_mutator_from_dict(mutator_dict)
            get_logger().debug(f"Applying mutator '{mutator.mutator_id}'.")
            img = mutator.mutate(img)

        return img

    def interpret_image(self, img: cv2.Mat, /) -> any:
        """
        Takes an image and runs the interpretation method defined in the corresponding feature class.
        Assumes that the feature class is present.

        Arguments:
            img: The image which should be passed to the intepretation method.

        Returns:
            The result of running the interpretation method on the image.
        """

        feature_class = self.get_feature_class()

        assert feature_class is not None

        interpretation_method_id = feature_class.get_data()["interpretation"]["method"]
        interpretation_method_config = feature_class.get_data()["interpretation"]["config"]

        assert isinstance(interpretation_method_id, str)
        assert isinstance(interpretation_method_config, dict)

        interpretation_method = load_interpretation_method(self._context, interpretation_method_id, interpretation_method_config)

        return interpretation_method.interpret(img, self._template_id, self.region_id)



ERR_INTERNAL = (1, "INTERNAL")

# IO errors
ERR_IO_INVALID_SUPERVISION_ENGINE = (101, "INVALID_IO_ENGINE")
ERR_IO_OPERATION_NOT_SUPPORTED_BY_DRIVER = (102, "OPERATION_NOT_SUPPORTED_BY_DRIVER")
ERR_IO_INVALID_PATH = (103, "INVALID_PATH")
ERR_IO_INVALID_IMAGE = (104, "INVALID_IMAGE")

# Matching errors
ERR_MATCHING_MATCH_COUNT_OUT_OF_BOUNDS = (201, "MATCH_COUNT_OUT_OF_BOUNDS")
ERR_MATCHING_INVALID_ENGINE_CONFIG = (202, "INVALID_ENGINE_CONFIG")

# Supervision errors
ERR_SUPERVISION_CORRESPONDENCE_NOT_FOUND = (301, "CORRESPONDENCE_NOT_FOUND")
ERR_SUPERVISION_INVALID_ENGINE_CONFIG = (302, "INVALID_ENGINE_CONFIG")

# Template errors
ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE = (401, "INVALID_SUPERVISION_ENGINE")
ERR_TEMPLATE_INVALID_MATCHING_ENGINE = (402, "INVALID_MATCHING_ENGINE")
ERR_TEMPLATE_ID_NOT_UNIQUE = (403, "ID_NOT_UNIQUE")
ERR_TEMPLATE_INVALID_KEYPOINT = (404, "INVALID_KEYPOINT")
ERR_TEMPLATE_INVALID_FEATURE = (405, "INVALID_FEATURE")
ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG = (406, "INVALID_CONCURRENCY_CONFIG")
ERR_TEMPLATE_INVALID_SYNTAX = (407, "INVALID_SYNTAX")
ERR_TEMPLATE_INVALID_FEATURE_CLASS = (408, "INVALID_FEATURE_CLASS")
ERR_TEMPLATE_INVALID_MUTATOR = (409, "INVALID_MUTATOR")
ERR_TEMPLATE_INVALID_INTERPRETATION = (410, "INVALID_INTERPRETATION")

# TODO: search for all raise statements in the entire project, and ensure that every call specifies arguments in accordance with a certain convention



# needed to not break type annotations if we are not in type checking mode
from __future__ import annotations

import os
import tempfile
from typing import TYPE_CHECKING, Dict, List, Union

import click
import cv2
import z3

from officialeye._internal.error.error import OEError
from officialeye._internal.error.errors.internal import ErrInternal
from officialeye._internal.error.errors.template import ErrTemplateIdNotUnique
from officialeye._internal.logger.singleton import get_logger

if TYPE_CHECKING:
    from officialeye._internal.io.driver import IODriver
    from officialeye._internal.template.template import Template


# TODO: move part of the Context class methods to IO driver

class Context:

    def __init__(self, manager, /, *, visualization_generation: bool = False):
        self._manager = manager

        self._io_driver: Union[IODriver, None] = None

        self._visualization_generation = visualization_generation

        self._export_counter = 1
        self._not_deleted_temporary_files: List[str] = []

        # keys: template ids
        # values: template
        self._loaded_templates: Dict[str, Template] = {}

        z3.set_param("parallel.enable", True)

    def visualization_generation_enabled(self) -> bool:
        return self._visualization_generation

    def get_io_driver(self) -> IODriver:

        if self._io_driver is None:
            raise ErrInternal(
                "while trying to access officialeye's IO driver.",
                "The present officialeye context does not have an IO Driver set."
            )

        return self._io_driver

    def set_io_driver(self, io_driver: IODriver, /):
        assert io_driver is not None
        self._io_driver = io_driver

    def add_template(self, template: Template, /):

        if template.template_id in self._loaded_templates:
            raise ErrTemplateIdNotUnique(
                f"while loading template '{template.template_id}'",
                "A template with the same id has already been loaded."
            )

        self._loaded_templates[template.template_id] = template

        try:
            template.validate()
        except OEError as err:
            # rollback the loaded template
            del self._loaded_templates[template.template_id]
            # reraise the cause
            raise err

    def get_template(self, template_id: str, /) -> Template:
        assert template_id in self._loaded_templates, "Unknown template id"
        return self._loaded_templates[template_id]

    def _allocate_file_name(self) -> str:
        file_name = "%03d.png" % self._export_counter
        self._export_counter += 1
        return file_name

    def allocate_file_for_export(self, /, *, file_name: str = "") -> str:

        if self._manager.export_directory is None:
            with tempfile.NamedTemporaryFile(prefix="officialeye_", suffix=".png", delete=False) as fp:
                fp.close()
            self._not_deleted_temporary_files.append(fp.name)
            return fp.name

        if file_name == "":
            file_name = self._allocate_file_name()

        return os.path.join(self._manager.export_directory, file_name)

    def export_image(self, img: cv2.Mat, /, *, file_name: str = "") -> str:
        export_file_path = self.allocate_file_for_export(file_name=file_name)
        cv2.imwrite(export_file_path, img)
        get_logger().info(f"Exported '{export_file_path}'.")
        return export_file_path

    def _export_and_show_image(self, img: cv2.Mat, /, *, file_name: str = ""):
        path = self.export_image(img, file_name=file_name)
        click.launch(path, locate=False)
        click.pause()

    def export_primary_image(self, img: cv2.Mat, /, *, file_name: str = ""):
        if get_logger().quiet_mode:
            # just save the image, do not export
            self.export_image(img, file_name=file_name)
        else:
            self._export_and_show_image(img, file_name=file_name)

    def _cleanup_temporary_files(self):
        while len(self._not_deleted_temporary_files) > 0:
            temp_file = self._not_deleted_temporary_files.pop()
            if os.path.exists(temp_file):
                os.unlink(temp_file)

    def dispose(self):
        self._cleanup_temporary_files()



from typing import Dict

from officialeye._internal.diffobject.difference_expansion import DiffObjectExpansion
from officialeye._internal.template.feature_class.const import IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID
from officialeye._internal.template.schema.schema import feature_class_object_specification


class FeatureClass:

    def __init__(self, manager, class_id: str, def_dict: Dict[str, any], /):
        self._manager = manager
        self.class_id = class_id
        self.is_inline = False

        self._data = {
            "abstract": False,
            "inherits": IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID,
            **def_dict
        }

    def is_global_base_class(self) -> bool:
        return self.class_id == IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID

    def get_parent_class(self):
        assert self.class_id != IMPLICIT_FEATURE_CLASS_BASE_INSTANCE_ID, "The global base class has no parent"
        parent_id = self._data["inherits"]
        return self._manager.get_class(parent_id)

    def is_abstract(self) -> bool:
        return self._data["abstract"]

    def get_features(self):
        """
        Generates all features that inherit this class (or belong to it directly)
        """

        template = self._manager.get_template()

        for cur_feature in template.features():
            cur_feature_class = cur_feature.get_feature_class()

            if cur_feature_class is None:
                continue

            # search in cur_feature's parents for the present class
            while not cur_feature_class.is_global_base_class():

                if cur_feature_class == self:
                    yield cur_feature
                    # we already found this class among the parents, no need to go deeper
                    break

                cur_feature_class = cur_feature_class.get_parent_class()

    def inline(self):
        """
        Computes all inherited attributes and applies diffs accordingly.
        The result gets cached so that they do not have to be computed every time.

        Raises: DiffObjectException
        """

        if self.is_inline:
            return

        class_parents_stack = []
        current_class = self

        while not current_class.is_global_base_class():
            class_parents_stack.append(current_class)
            current_class = current_class.get_parent_class()

        if len(class_parents_stack) == 0:
            class_parents_stack = [self]

        expansion = DiffObjectExpansion(feature_class_object_specification)

        while len(class_parents_stack) > 0:
            ancestor = class_parents_stack.pop()
            expansion.add(ancestor._data)

            if ancestor.is_abstract():
                ancestor_inlined = expansion.get_current_partial_object()
            else:
                ancestor_inlined = expansion.get_full_object()

            ancestor.is_inline = True
            ancestor._data = ancestor_inlined

        assert self.is_inline

    def get_data(self) -> Dict[str, any]:
        return self._data
